#include <stdlib.h>
#include <sys/time.h>
#include "mpi.h"
#include "mpe.h"
#include "p_m_gen.h"

#define DEBUG 0
#define POINTS 0
#define PROFILE 0
#define LOG 1


#define USE_WORKER 0
#define OLD_MPICH 0
#define MASTER_PROC 0

/* shading techniques: */
#define MAXITER_SHADE 1
/* (n==maxiter) */
#define EVEN_SHADE    2
/* (n==maxiter || !(n%2)) */


/* defaults: */
#define DEF_rmin     -2.0
#define DEF_rmax      2.0
#define DEF_imin     -2.0
#define DEF_imax      2.0
#define DEF_width     500
#define DEF_height    500
#define DEF_boundary  2.0
#define DEF_maxiter   100
#define DEF_bw        0
#define DEF_breakout  0
#define DEF_tasksize  1
#define DEF_logfile   "p_m3.log"
#define DEF_useinf    0
#define DEF_inputfile "p_m3.in"
#define DEF_randomize 0
#define DEF_numColors 16
#define DEF_colScale  7

#define INIT_QUEUE_SIZE 100
/* initial size of the rect-to-do queue */
#define MAX_RECT_PASSED 4
/* the greatest number of rect definitions a slave can pass to the master */

typedef struct {
  int l, r, t, b;
} rect;

typedef struct {
  int head, tail, size;
  rect *r;
} rect_queue;

/* messge tags: */
/*    master to slave: */
#define READY_TO_START   42
#define READY_FOR_MORE   43
#define ADD2Q            44
#define RECTS_TO_ENQUEUE 45
/*    slave to master: */
#define ASSIGNMENT       46
#define ALL_DONE         47


/* logfile events */
#define S_COMPUTE 10
#define E_COMPUTE 11
#define S_DRAW_BLOCK 12
#define E_DRAW_BLOCK 13
#define S_WAIT_FOR_MESSAGE 14
#define E_WAIT_FOR_MESSAGE 15
#define S_DRAW_RECT 16
#define E_DRAW_RECT 17
#define S_DRAW_CHUNK 18
#define E_DRAW_CHUNK 19

#define Q_EMPTY (rect_q.head == rect_q.tail)

#define ITER2COLOR( iter ) ( ((iter)==maxiter) ? (0) : ((iter) % numColors) )

#define IS_SET(z) ( \
  (bw==MAXITER_SHADE) ? ((z)==maxiter) : \
  (bw==EVEN_SHADE)    ? ((z)==maxiter || !((z)%2)) : (z))

#define BORDER_LEN( rect ) ( \
  ((rect).r-(rect).l) ? \
    ((rect).b-(rect).t) ? \
      (2 * ((rect).r-(rect).l+(rect).b-(rect).t) ) \
    : \
      ((rect).r - (rect).l + 1) \
  : \
    ((rect).b-(rect).t) ? \
      ((rect).b - (rect).t + 1) \
    : \
      1 \
)


#define CHECK_Q_SIZE { \
  if (rect_q.head == rect_q.tail+1 ||                                            /* if the queue is full */ \
      !rect_q.head && rect_q.tail == rect_q.size - 1) { \
    rect_q.r = (rect *) realloc( rect_q.r, sizeof( rect ) * rect_q.size * 2 );	 /* get a bigger queue */ \
    if (rect_q.tail < rect_q.head) { \
      memcpy( rect_q.r + rect_q.size, rect_q.r, rect_q.tail * sizeof( rect ) );	 /* copy over any data that needs to be moved */ \
      rect_q.tail += rect_q.size; \
    } \
    if (randomize && randomPt<rect_q.head) { \
      randomPt += rect_q.size; \
    } \
    rect_q.size *= 2; \
  } \
}


#define RECT_ASSIGN( rect, w, x, y, z ) { \
  (rect).l = (w); (rect).r = (x); (rect).t = (y); (rect).b = (z); }

#define ENQUEUE(w, x, y, z) { \
  CHECK_Q_SIZE; \
  RECT_ASSIGN( rect_q.r[rect_q.tail], (w), (x), (y), (z) ); \
  if (++rect_q.tail == rect_q.size) rect_q.tail = 0; \
}

/*
fprintf( stderr, "head moved to %d\n", rect_q.head ); \
fprintf( stderr, "RANDOMIZE!!\nhead = %d, tail = %d, size = %d\n", rect_q.head, rect_q.tail, rect_q.size ); \
fprintf( stderr, "swap %d and %d\n", i, j ); \
*/
#define DEQUEUE { \
  if (++rect_q.head == rect_q.size) rect_q.head = 0; \
  if (randomize && rect_q.head == randomPt) { \
    int i, j, numItems; \
    rect temp; \
fprintf( stderr, "RANDOMIZE!!\nhead = %d, tail = %d, size = %d\n", rect_q.head, rect_q.tail, rect_q.size ); \
    numItems = (rect_q.tail<rect_q.head) \
      ? rect_q.size-rect_q.head + rect_q.tail \
      : rect_q.tail - rect_q.head; \
    for (i=rect_q.head; i!=rect_q.tail; i++) { \
      j = random() % numItems + rect_q.head; \
      if (j>=rect_q.size) j-=rect_q.size; \
      temp = rect_q.r[j]; \
      rect_q.r[j] = rect_q.r[i]; \
      rect_q.r[i] = temp; \
      if (i==rect_q.size-1) { \
	i = -1; \
      } \
    } \
    randomPt = rect_q.tail; \
  } \
}

#define ENQUEUE_STRUCT(rect) { \
  CHECK_Q_SIZE; \
  rect_q.r[rect_q.tail] = (rect); \
  if (++rect_q.tail == rect_q.size) rect_q.tail = 0; \
}

#define SPLIT_RECT( r ) { \
  xsplit = (r.r-r.l)>>1; \
  ysplit = (r.b-r.t)>>1; \
  RECT_ASSIGN( rectBuf[0], r.l+1,        r.l+xsplit, r.t+1,       r.t+ysplit ); \
  RECT_ASSIGN( rectBuf[1], r.l+xsplit+1, r.r-1,      r.t+1,        r.t+ysplit ); \
  RECT_ASSIGN( rectBuf[2], r.l+1,        r.l+xsplit, r.t+ysplit+1, r.b-1 ); \
  RECT_ASSIGN( rectBuf[3], r.l+xsplit+1, r.r-1,      r.t+ysplit+1, r.b-1 ); \
  xsplit = 4;									 /* number of rects sent */ \
  MPI_Send( &xsplit, 1, MPI_INT, MASTER_PROC, ADD2Q, MPI_COMM_WORLD );		 /* tell the master process we're send 4 rects */ \
  MPI_Send( rectBuf, sizeof( rect ) * 4, MPI_BYTE, MASTER_PROC,	                 /* send the 4 rects */ \
	    RECTS_TO_ENQUEUE, MPI_COMM_WORLD ); \
}



#if DEBUG
#define CALC_POINT_AND_CHECK { \
  *datPtr = MbrotCalcIter( re, im );					 /* calculate iterations */ \
  fprintf( outf, " (%d %d) (%lf,%lf) %d\n", x, y, re, im, *datPtr ); \
  if (isContinuous && *borderData != *datPtr) {				 /* if first aberration found, */ \
    fprintf( outf, "different " ); \
    SPLIT_RECT( r ); \
    isContinuous = 0; \
  }\
  datPtr++; \
}
#else
#define CALC_POINT_AND_CHECK { \
  *datPtr = MbrotCalcIter( re, im );					 /* calculate iterations */ \
  if (isContinuous && *borderData != *datPtr) {				 /* if first aberration found, */ \
    SPLIT_RECT( r ); \
    isContinuous = 0; \
  }\
  datPtr++; \
}
#endif

#define CALC_POINT { \
  *datPtr = MbrotCalcIter( re, im ); \
  datPtr++; \
}


void ProcessArgs( argc, argv, width, height, rmin, rmax, imin, imax,
		  boundary, maxiter, bw, breakout, tasksize,
		  logfile, randomize, numColors)
int argc, *width, *height, *maxiter, *bw,
    *breakout, *tasksize, *randomize, *numColors;
double *boundary;
NUM *rmin, *rmax, *imin, *imax;
char **argv, **logfile;
{
  int argNum;
  double temp;

  for (argNum=1; argNum < argc; argNum++) {
#if DEBUG
fprintf( stderr, "argv[%d] = %s\n", argNum, argv[argNum] );
#endif
    if (!strcmp( argv[argNum], "-boundary" )) {
      sscanf( argv[++argNum], "%lf", boundary );
      
    } else if (!strcmp( argv[argNum], "-rmin" )) {
      sscanf( argv[++argNum], "%lf", &temp );
      NUM_PTR_ASSIGN( rmin, DBL2NUM( temp ) );
	
    } else if (!strcmp( argv[argNum], "-rmax" )) {
      sscanf( argv[++argNum], "%lf", &temp );
      NUM_PTR_ASSIGN( rmax, DBL2NUM( temp ) );
	
    } else if (!strcmp( argv[argNum], "-imin" )) {
      sscanf( argv[++argNum], "%lf", &temp );
      NUM_PTR_ASSIGN( imin, DBL2NUM( temp ) );
	
    } else if (!strcmp( argv[argNum], "-imax" )) {
      sscanf( argv[++argNum], "%lf", &temp );
      NUM_PTR_ASSIGN( imax, DBL2NUM( temp ) );
	
    } else if (!strcmp( argv[argNum], "-maxiter" )) {
      sscanf( argv[++argNum], "%d", maxiter );

    } else if (!strcmp( argv[argNum], "-width" )) {
      sscanf( argv[++argNum], "%d", width );

    } else if (!strcmp( argv[argNum], "-height" )) {
      sscanf( argv[++argNum], "%d", height );

    } else if (!strcmp( argv[argNum], "-bw" )) {
      argNum++;
      if (!strcmp( argv[argNum], "even" )) {
	*bw = EVEN_SHADE;
      } else if (!strcmp( argv[argNum], "maxiter" )) {
	*bw = MAXITER_SHADE;
      } else {
	printf( "Unknown shade type \"%s\".  Using \"even\".\n",
	        argv[argNum] );
	*bw = EVEN_SHADE;
      }

    } else if (!strcmp( argv[argNum], "-breakout" )) {
      sscanf( argv[++argNum], "%d", breakout );

    } else if (!strcmp( argv[argNum], "-tasksize" )) {
      sscanf( argv[++argNum], "%d", tasksize );

    } else if (!strcmp( argv[argNum], "-colors" )) {
      sscanf( argv[++argNum], "%d", numColors );

    } else if (!strcmp( argv[argNum], "-l" )) {
      *logfile = argv[++argNum];

    } else if (!strcmp( argv[argNum], "-randomize" )) {
      *randomize = 1;

    } else if (!strcmp( argv[argNum], "-h" )) {
      printf( "Options recognized by %s:\n", argv[0] );
      printf( "(defaults are in parentheses () )\n" );
      printf( "\t-o         (<stdout>) output file\n" );
      printf( "\t-l         (\"%s\") name of log file\n", DEF_logfile );
      printf( "\t-width     (%d) width of computed area in points\n", DEF_width );
      printf( "\t-height    (%d) height of computed area in points\n", DEF_height );
      printf( "\t-boundary  (%.1lf) boundary point for M-set computation algorithm\n", DEF_boundary );
      printf( "\t-maxiter   (%d) maximum # of iterations for M-set compuptation algorithm\n", DEF_maxiter );
      printf( "\t-rmin      (%.2lf) minimum real coordinate of computed area\n", DEF_rmin );
      printf( "\t-rmax      (%.2lf) maximum real coordinate of computed area\n", DEF_rmax );
      printf( "\t-imin      (%.2lf) minimum imaginary coordinate of computed area\n", DEF_imin );
      printf( "\t-imax      (%.2lf) maximum imaginary coordinate of computed area\n", DEF_imax );
      printf( "\t-breakout  (%d) maximum length or width rectangle to subdivide\n", DEF_breakout );
      printf( "\t-colors    (%d) number of colors to request\n", DEF_numColors );
      printf( "\t-randomize (%sset) compute regions in as random of order as possible\n", DEF_randomize ? "" : "not " );
/*
      printf( "\t-tasksize  (%d) approximate number of pixels to assign a slave\n", DEF_tasksize );
      printf( "\t                process after each request for work\n" );
*/
      printf( "\t-bw        (%sset) draw in black and white instead of color\n", DEF_bw ? "" : "not " );
      printf( "\t               maxiter - fill where 'iter' = maxiter\n" );
      printf( "\t               even - fill where 'iter' is even\n" );
      exit( 0 );
    }
  }
}



void DrawRect( outf, graph, r, borderData, maxiter, bw, numColors )
FILE *outf;
MPE_XGraph *graph;
rect r;
int *borderData, maxiter, bw, numColors;
{
  int x, y;

  if (bw) {
    for (y=r.t, x=r.l; x<=r.r; x++) {
#if POINTS
fprintf( outf, "draw %d at %d %d\n", *borderData, x, y );
#endif
      if (IS_SET(*borderData)) {
	MPE_DrawPoint( graph, x, y, 1 );
      }   
      borderData++;
    }
    for (x=r.r, y=r.t+1; y<=r.b; y++) {
#if POINTS
fprintf( outf, "draw %d at %d %d\n", *borderData, x, y );
#endif
      if (IS_SET(*borderData)) {
	MPE_DrawPoint( graph, x, y, 1 );
      }   
      borderData++;
    }
    if (r.r-r.l && r.b-r.t) {
      for (y=r.b, x=r.r-1;x>=r.l; x--) {
#if POINTS
fprintf( outf, "draw %d at %d %d\n", *borderData, x, y );
#endif
	if (IS_SET(*borderData)) {
	  MPE_DrawPoint( graph, x, y, 1 );
	}   
	borderData++;
      }
      for (x=r.l, y=r.b-1; y>r.t; y--) {
#if POINTS
fprintf( outf, "draw %d at %d %d\n", *borderData, x, y );
#endif
	if (IS_SET(*borderData)) {
	  MPE_DrawPoint( graph, x, y, 1 );
	}   
	borderData++;
      }
    }
  } else { /* in color */
    for (y=r.t, x=r.l; x<=r.r; x++) {
#if POINTS
fprintf( outf, "draw %d at %d %d\n", *borderData, x, y );
#endif
      MPE_DrawPoint( graph, x, y, ITER2COLOR(*borderData) );
      borderData++;
    }
    for (x=r.r, y=r.t+1; y<=r.b; y++) {
#if POINTS
fprintf( outf, "draw %d at %d %d\n", *borderData, x, y );
#endif
      MPE_DrawPoint( graph, x, y, ITER2COLOR(*borderData) );
      borderData++;
    }
    if (r.r-r.l && r.b-r.t) {
      for (y=r.b, x=r.r-1;x>=r.l; x--) {
#if POINTS
fprintf( outf, "draw %d at %d %d\n", *borderData, x, y );
#endif
	MPE_DrawPoint( graph, x, y, ITER2COLOR(*borderData) );
	borderData++;
      }
      for (x=r.l, y=r.b-1; y>r.t; y--) {
#if POINTS
fprintf( outf, "draw %d at %d %d\n", *borderData, x, y );
#endif
	MPE_DrawPoint( graph, x, y, ITER2COLOR(*borderData) );
	borderData++;
      }
    }
  }
  
  MPE_Update( graph );
}


void DrawBlock( outf, graph, r, iter, maxiter, bw, numColors )
MPE_XGraph *graph;
rect r;
int iter, maxiter, bw, numColors;
{
  int color, x, y;

#if POINTS
for (y=r.t; y<=r.b; y++) {
  for (x=r.l; x<=r.r; x++) {
    fprintf( outf, "draw %d at %d %d\n", iter, x, y );
  }
}
#endif
  if (bw) {
    if (IS_SET( iter )) {
      MPE_FillRectangle( graph, r.l, r.t, r.r-r.l+1, r.b-r.t+1, 1 );
    }
  } else {  /* color */
    color = ITER2COLOR(iter);
    MPE_FillRectangle( graph, r.l, r.t, r.r-r.l+1, r.b-r.t+1, color );
  }
  
  MPE_Update( graph );
}


void DrawChunk( outf, graph, chunkField, r, maxiter, bw, numColors )
MPE_XGraph *graph;
int *chunkField, maxiter, bw, numColors;
rect r;
{
  int a, b;

  if (bw) {
    for (b=r.t; b<=r.b; b++) {
      for (a=r.l; a<=r.r; a++) {
#if POINTS
fprintf( outf, "draw %d at %d %d\n", *chunkField, a, b );
#endif
	if (IS_SET(*chunkField)) {
	  MPE_DrawPoint( graph, a, b, 1 );
	}
	chunkField++;
      }
    }
  } else {
    for (b=r.t; b<=r.b; b++) {
      for (a=r.l; a<=r.r; a++) {
#if POINTS
fprintf( outf, "draw %d at %d %d\n", *chunkField, a, b );
#endif
	MPE_DrawPoint( graph, a, b, ITER2COLOR( *chunkField ) );
	chunkField++;
      }
    }
  }

  MPE_Update( graph );
}


SumInts( borderData, r, isContinuous )
int *borderData, isContinuous;
rect r;
{
  int sum=0, npoints;
  if (isContinuous) return ((r.r-r.l+1) * (r.b-r.t+1) * *borderData);
  
  for (npoints = BORDER_LEN( r ); npoints; npoints--) {
    sum += *borderData++;
  }
  return sum;
}

SumChunkInts( chunkField, r )
int *chunkField;
rect r;
{
  int sum=0, npoints;
  for (npoints = (r.r-r.l+1)*(r.b-r.t+1); npoints; npoints--) {
    sum += *chunkField++;
  }
  return sum;
}


void CopyPointToField( iterField, width, height, data, r )
int *iterField, width, height, data;
rect r;
{
  int i, j, *writePtr;

  writePtr = iterField + width*r.t + r.l;
  for (j=r.t; j<=r.b; j++) {
    for (i=r.l; i<=r.r; i++) {
      *writePtr++ = data;
    }
    writePtr += width - r.r + r.l - 1;
  }
}

void Master_Proc( width, height, maxiter, boundary,
		  rmin, rmax, imin, imax, tasksize,
		  randomize)
int width, height, maxiter, tasksize, randomize;
NUM rmin, rmax, imin, imax;
double boundary;
{
  rect *assigList, recvRectBuf[MAX_RECT_PASSED];
  /* assigList - list of what rectangles have been assigned to each process */
  /* recvRectBuf - when a slave process tells the master that some rectangles
     need to be calculated, the rectangle definitions are temporarily stored
     here */

  int inProgress, nidle, *idleList, splitPt, np, data,
      mesgTag, firstToEnqueue, procNum, i, randomPt;
  /* inProgress - number of rectangles currently under construction */
  /* nidle - number of idle processes */
  /* idleList - list of process ranks that are idle */
  /* splitPt - point at which to split the Mandelbrot set--can't have the
     border algorithm enclosing the whole set */
  /* np - number of processes */
  /* data - one int returned by slave processes; different uses */
  /* mesgTag - the tag of the received message */
  /* firstToEnqueue - when receiving a bunch of rectangles to enqueue, the
     master process may send some off to idle processes right away before
     queueing them.  firstToEnqueue points to the first one that is actually
     queued */
  /* procNum - rank of the process that sent the last message */
  /* randomPt - index+1 of the last item in the queue that has not been
                placed in random order */

  rect_queue rect_q;
  /* queue of rectangles to calculate */

  MPI_Status mesgStatus;

  struct timeval time;

  MPI_Comm_size( MPI_COMM_WORLD, &np );
  gettimeofday( &time, 0 );			     /* initialize the random number generator for the -randomize option */
  srandom( (int)time.tv_usec );
#if LOG
  MPE_Describe_state( S_COMPUTE, E_COMPUTE, "Compute", "blue:gray" );
  MPE_Describe_state( S_DRAW_BLOCK, E_DRAW_BLOCK, "Draw block", "yellow:gray3" );
  MPE_Describe_state( S_DRAW_RECT, E_DRAW_RECT, "Draw border", "green:light_gray" );
  MPE_Describe_state( S_WAIT_FOR_MESSAGE, E_WAIT_FOR_MESSAGE, "Wait for message", "red:black" );
  MPE_Describe_state( S_DRAW_CHUNK, E_DRAW_CHUNK, "Draw Chunk", "steelBlue:2x2" );
#endif
  assigList  = (rect *) malloc( (np) * sizeof( rect ) );
  idleList   = (int *)  malloc( (np) * sizeof( rect ) );

  nidle = inProgress = 0;
  rect_q.head = rect_q.tail = 0;						 /* create the queue */
  rect_q.size =  INIT_QUEUE_SIZE;
  rect_q.r = (rect *) malloc( rect_q.size * sizeof( rect ) );
  randomPt = 1;

  if (imin<0 && imax>0) {							 /* might have to split the set */
    splitPt = height + NUM2INT( NUM_DIV( NUM_MULT( imin, INT2NUM( height ) ),
					 NUM_SUB( imax, imin ) ) );

    ENQUEUE( 0, width-1, 0, splitPt-1 );
    ENQUEUE( 0, width-1, splitPt, height-1 );
  } else {
    ENQUEUE( 0, width-1, 0, height-1 );
  }


  while (inProgress || (rect_q.head != rect_q.tail) ) {				 /* while someone's working, and the q is !empty */
#if LOG
    MPE_Log_event( S_WAIT_FOR_MESSAGE, 0, 0 );
#endif
    MPI_Recv( &data, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG,			 /* receive a message */
	      MPI_COMM_WORLD, &mesgStatus );
#if LOG
    MPE_Log_event( E_WAIT_FOR_MESSAGE, 0, 0 );
#endif
    procNum = mesgStatus.MPI_SOURCE;
    mesgTag = mesgStatus.MPI_TAG;

#if DEBUG
fprintf( stderr, "Master receives %d from %d\n", mesgTag, procNum );
#endif

    switch (mesgTag) {

    case READY_TO_START:
      inProgress++;
    case READY_FOR_MORE:
      if (Q_EMPTY) {								 /* if the queue is empty, */
	idleList[nidle++] = procNum;						 /* remember this process was left idle */
	inProgress--;
      } else {
	MPI_Send( rect_q.r + rect_q.head, sizeof( rect ), MPI_BYTE,		 /* otherwise, assign it the next in the queue */
		 procNum, ASSIGNMENT, MPI_COMM_WORLD );
	assigList[procNum] = rect_q.r[rect_q.head];				 /* remember which rect this process is on */
#if DEBUG
fprintf( stderr, "Master assigns (%d %d %d %d) to %d\n", assigList[procNum].l,
	 assigList[procNum].r, assigList[procNum].t, assigList[procNum].b,
	 procNum );
#endif
	DEQUEUE;
      }
      break;

    case ADD2Q:								      /* slave is posting more rectangles to be queued */
      MPI_Recv( recvRectBuf, data * sizeof( rect ), MPI_BYTE, procNum,		 /* get rect definitions */
	        RECTS_TO_ENQUEUE, MPI_COMM_WORLD, &mesgStatus );
      firstToEnqueue = 0;
      while (nidle && firstToEnqueue < data) {					 /* if processes are idle, */
	nidle--;
	assigList[idleList[nidle]] = recvRectBuf[firstToEnqueue];		 /* remember which rect this process is on */
	MPI_Send( recvRectBuf + firstToEnqueue, sizeof( rect ), MPI_BYTE,	 /* give them something to do */
		  idleList[nidle], ASSIGNMENT, MPI_COMM_WORLD );
#if DEBUG
fprintf( stderr, "Master assigns (%d %d %d %d) to %d\n", (recvRectBuf[firstToEnqueue]).l,
	 (recvRectBuf[firstToEnqueue]).r, (recvRectBuf[firstToEnqueue]).t,
	 (recvRectBuf[firstToEnqueue]).b, idleList[nidle] );
#endif
	inProgress++; firstToEnqueue++;
      }
      for (; firstToEnqueue<data; firstToEnqueue++) {
	ENQUEUE_STRUCT( recvRectBuf[firstToEnqueue] );
      }
      break;
    }
  }

#if DEBUG
fprintf( stderr, "All done\n" );
#endif

  for (i=1; i<np; i++) {							 /* tell everyone to exit */
    MPI_Send( rect_q.r, sizeof( rect ), MPI_BYTE, i, ALL_DONE, MPI_COMM_WORLD );
  }
}



void Slave_Proc( graph, width, height, maxiter, boundary,
		 rmin, rmax, imin, imax, bw,
		 breakout, tasksize, numColors )
MPE_XGraph *graph;
int width, height, maxiter, bw, breakout, tasksize, numColors;
double boundary;
NUM rmin, rmax, imin, imax;
{
#if PROFILE
  int prof_nchunks=0, prof_npoints=0, prof_niters=0;
  double prof_calcTime=0, prof_thisStartTime;
#endif
  int x, y, working, isContinuous, *borderData, *datPtr, mesgTag, myid;
  /* x, y - integer counters for the point being calculated */
  /* working - whether this process is still working or has been retired */
  /* borderData - storage for the values calculated */
  /* datPtr - pointer into borderData */
  /* mesgTag - what type of message was just received */

  int xsplit, ysplit, *chunkField;
  /* where to split when dividing up the job, rounded down */

  NUM re, im, rstep, istep;
  /* re, im - the complex point being calculated */
  /* rstep, istep - distance between complex points that are calculated */

  rect r, rectBuf[MAX_RECT_PASSED];
  /* r - the rectangle being calculated */

  MPI_Status mesgStatus;

  char outfilename[20];
  FILE *outf;
  /* per-process output file for debugging and profiling info */

  MPI_Comm_rank( MPI_COMM_WORLD, &myid );

#if DEBUG || PROFILE || POINTS
  sprintf( outfilename, "p_m3_proc%d.out", myid );
  outf = fopen( outfilename, "w" );
#endif

  MPI_Send( &myid, 1, MPI_INT, MASTER_PROC, READY_TO_START, MPI_COMM_WORLD );
  
  working = isContinuous = 1;
  NUM_ASSIGN( rstep,  NUM_DIV( NUM_SUB( rmax, rmin ), INT2NUM( width-1 ) ) );
  NUM_ASSIGN( istep,  NUM_DIV( NUM_SUB( imin, imax ), INT2NUM( height-1 ) ) );
  borderData = (int *) malloc( sizeof( int ) * 2 * (height + width) );
  if (breakout) chunkField = (int *) malloc( sizeof( int ) * breakout * breakout );
  Mbrot_Settings( boundary, maxiter );

  while (working) {
#if LOG
    MPE_Log_event( S_WAIT_FOR_MESSAGE, 0, 0 );
#endif
    MPI_Recv( &r, sizeof( rect ), MPI_BYTE, MASTER_PROC, MPI_ANY_TAG, MPI_COMM_WORLD,      /* get command from master process */
	      &mesgStatus );
#if LOG
    MPE_Log_event( E_WAIT_FOR_MESSAGE, 0, 0 );
#endif
    mesgTag = mesgStatus.MPI_TAG;

#if DEBUG
fprintf( outf, "receive: %d\n", mesgTag);
#endif

    switch (mesgTag) {

    case ASSIGNMENT:								 /* new rectangle to compute */
#if DEBUG
fprintf( outf, "Assigned (%d %d %d %d)\n", r.l, r.r, r.t, r.b );
#endif
#if PROFILE
      prof_thisStartTime = MPE_Ptime();
#endif
#if LOG
      MPE_Log_event( S_COMPUTE, 0, 0 );
#endif
      if (r.b-r.t<breakout || r.r-r.l<breakout) {                                /* if smaller than breakout, compute directly */
#if DEBUG
fprintf( outf, "Compute directly\n" );
#endif
	CalcField( outf, MbrotCalcIter, chunkField, rmin, rmax, imin, imax,	 /* compute the field */
		   0, width-1, 0, height-1, r.l, r.r, r.t, r.b );
#if PROFILE
      prof_calcTime += MPE_Ptime()-prof_thisStartTime;
      prof_npoints  += (r.r-r.l+1) * (r.b-r.t+1);
      prof_niters   += SumChunkInts( chunkField, r );
      prof_nchunks++;
#endif
#if LOG
	MPE_Log_event( E_COMPUTE, 0, 0 );
        MPE_Log_event( S_DRAW_CHUNK, 0, 0 );
#endif
	DrawChunk( outf, graph, chunkField, r, maxiter, bw, numColors );	 /* draw it */
#if LOG
        MPE_Log_event( E_DRAW_CHUNK, 0, 0 );
#endif

      } else {									 /* otherwise, compute the boundary */
	NUM_ASSIGN( re, COORD2CMPLX( rmin, rmax, 0, width-1,  r.l) );		 /* set the complex points */
	NUM_ASSIGN( im, COORD2CMPLX( imax, imin, 0, height-1, r.t) );
	datPtr = borderData;
	*datPtr++ = MbrotCalcIter( re, im );					 /* calculate first point */
#if DEBUG
fprintf( outf, "First point: %d\n", *borderData );
#endif
	if (r.b-r.t>1 && r.r-r.l>1) {			/* if there's a chance to subdivide, */
	  isContinuous=1;
	  for (y=r.t, x=r.l+1; x<=r.r; x++) {		/* loop through the top row (first point done) */
	    NUM_ASSIGN( re, NUM_ADD( re, rstep ) );	/* move to next complex coord */
	    CALC_POINT_AND_CHECK;
	  }
	  for (x=r.r, y=r.t+1; y<=r.b; y++) {		/* loop through the right side */
	    NUM_ASSIGN( im, NUM_ADD( im, istep ) );	/* move to next complex coord */
	    CALC_POINT_AND_CHECK;
	  }
	  for (y=r.b, x=r.r-1; x>=r.l; x--) {		/* loop through the bottom row */
	    NUM_ASSIGN( re, NUM_SUB( re, rstep ) );	/* move to next complex coord */
	    CALC_POINT_AND_CHECK;
	  }
	  for (x=r.l, y=r.b-1; y>r.t; y--) {		/* loop through the left side */
	    NUM_ASSIGN( im, NUM_SUB( im, istep ) );	/* move to next complex coord */
	    CALC_POINT_AND_CHECK;
	  }
	} else {						/* if there's no chance to subdivide, don't insert the checks */
	  isContinuous=0;
	  for (y=r.t, x=r.l+1; x<=r.r; x++) {		/* loop through the top row */
	    NUM_ASSIGN( re, NUM_ADD( re, rstep ) );	/* move to next complex coord */
	    CALC_POINT
	  }
	  for (x=r.r, y=r.t+1; y<=r.b; y++) {		/* loop through the right side */
	    NUM_ASSIGN( im, NUM_ADD( im, istep ) );	/* move to next complex coord */
	    CALC_POINT;
	  }
	  if (r.r-r.l && r.b-r.t) {			/* only do the opposite sides if >1 row and >1 column */
	    for (y=r.b, x=r.r-1; x>=r.l; x--) {		/* loop through the bottom row */
	      NUM_ASSIGN( re, NUM_SUB( re, rstep ) );	/* move to next complex coord */
	      CALC_POINT;
	    }
	    for (x=r.l, y=r.b-1; y>r.t; y--) {		/* loop through the left side */
	      NUM_ASSIGN( im, NUM_SUB( im, istep ) );	/* move to next complex coord */
	      CALC_POINT;
	    }
	  }
	}	/* if (r.b-r.t>1 || r.r-r.l>1) */
#if LOG
	MPE_Log_event( E_COMPUTE, 0, 0 );
#endif

#if PROFILE
      prof_calcTime += MPE_Ptime()-prof_thisStartTime;
      prof_npoints  += (r.r-r.l+1) * (r.b-r.t+1);
      prof_niters   += SumInts( borderData, r, isContinuous );
      prof_nchunks++;
#endif

	if (isContinuous) {
#if DEBUG
fprintf( outf, "Drawing block\n");
#endif
#if LOG
          MPE_Log_event( S_DRAW_BLOCK, 0, 0 );
#endif
          DrawBlock( outf, graph, r, *borderData, maxiter, bw, numColors );
#if LOG
          MPE_Log_event( E_DRAW_BLOCK, 0, 0 );
#endif
        } else {
#if DEBUG
fprintf( outf, "Drawing border\n");
#endif
#if LOG
          MPE_Log_event( S_DRAW_RECT, 0, 0 );
#endif
          DrawRect( outf, graph, r, borderData, maxiter, bw, numColors );
#if LOG
          MPE_Log_event( E_DRAW_RECT, 0, 0 );
#endif
        }
      }	/* else !breakout */
      MPI_Send( &working, 1, MPI_INT, MASTER_PROC, READY_FOR_MORE, MPI_COMM_WORLD );
      break;									 /* end if case ASSIGNMENT: */

    case ALL_DONE:
      working=0;
      break;

    } /* end of switch */
  } /* end of while (working) */

#if PROFILE
  fprintf( outf, "[%d] info from proc %d:\n", myid, myid );
  fprintf( outf, "[%d] # chunks:     %d\n", myid, prof_nchunks );
  fprintf( outf, "[%d] # points:     %d\n", myid, prof_npoints );
  fprintf( outf, "[%d] # iterations: %d\n", myid, prof_niters );
  fprintf( outf, "[%d] avg. iter/point: %.1lf\n", myid, (double)prof_niters/prof_npoints );
  fprintf( outf, "[%d] total calculation time: %.20lf sec.\n", myid, prof_calcTime );
  fprintf( outf, "[%d] avg. sec/chunk:         %.20lf sec.\n", myid, prof_calcTime/prof_nchunks );
  fprintf( outf, "[%d] avg. sec/iter:          %.20lf sec.\n", myid, prof_calcTime/prof_niters );
#endif

}




#if USE_WORKER
worker( argc, argv )
#else
main( argc, argv )
#endif
int argc;
char **argv;
{
  int myid, np, str_len;
  /* str_len - length of DISPLAY variable */

  MPE_XGraph *graph;

  char *display, *logfile = DEF_logfile;
  /* string form of the DISPLAY variable */
  /* logfile that is to be written to */

  MPE_Color *colorArray;

  double boundary=DEF_boundary;
  /* boundary - boundary point in Mandelbrot set calculation sequence */

  NUM rmin, rmax, imin, imax;
  /* rmin, rmax, imin, imax - the area to be calculated */

  int maxiter=DEF_maxiter, width=DEF_width, height=DEF_height,
      bw=DEF_bw, breakout=DEF_breakout, randomize=DEF_randomize,
      tasksize=DEF_tasksize, numColors=DEF_numColors;
  /* maxiter - maximum # of iterations for mbrot calculation */
  /* width, height - dimensions of the whole field to be calculated */
  /* bw - whether we're on a B&W display, and when to fill pixels
          if we are - MAXITER_SHADE or EVEN_SHADE */
  /* breakout - when subdividing, if either the width or the height of
     a region to be calculated drops to or below 'breakout', the region
     will be calculated outright */
  /* tasksize - maximum number of jobs to give a child process with each
     request.  This is to cut down on the messages being passed around. */
  /* randomize - whether to shuffle the queue up whenever possible */
  /* numColors - the number of colors to request from the display */

#if OLD_MPICH
  MPI_Init( &argc, argv );
#else
  MPI_Init( &argc, &argv );
#endif


#if LOG
  MPE_Initlog();
#endif
  MPI_Comm_rank( MPI_COMM_WORLD, &myid );
  MPI_Comm_size( MPI_COMM_WORLD, &np );
  
  NUM_ASSIGN( rmin, DBL2NUM( DEF_rmin ) );
  NUM_ASSIGN( rmax, DBL2NUM( DEF_rmax ) );
  NUM_ASSIGN( imin, DBL2NUM( DEF_imin ) );
  NUM_ASSIGN( imax, DBL2NUM( DEF_imax ) );

  if (myid==0) {
    ProcessArgs( argc, argv, &width, &height, &rmin, &rmax, &imin, &imax,
		 &boundary, &maxiter, &bw, &breakout, &tasksize,
		 &logfile, &randomize, &numColors);
    display = getenv( "DISPLAY" );
    if (!display) {
      printf ("DISPLAY environment variable not set.  Exiting.\n\n");
      exit( -1 );
    }
    if (!isalpha(*display)) {
      printf ("hostname not in DISPLAY variable, please add.\n\n");
      exit( -1 );
    }
    str_len = strlen( display ) + 1;
    MPI_Bcast( &str_len, 1, MPI_INT, 0, MPI_COMM_WORLD );
    MPI_Bcast( display, str_len, MPI_CHAR, 0, MPI_COMM_WORLD );
    str_len = strlen( logfile ) + 1;
    MPI_Bcast( &str_len, 1, MPI_INT, 0, MPI_COMM_WORLD );
    MPI_Bcast( logfile, str_len, MPI_CHAR, 0, MPI_COMM_WORLD );
  } else {
    MPI_Bcast( &str_len, 1, MPI_INT, 0, MPI_COMM_WORLD );
    display = (char *) malloc( sizeof( char ) * str_len );
    MPI_Bcast( display, str_len, MPI_CHAR, 0, MPI_COMM_WORLD );
    MPI_Bcast( &str_len, 1, MPI_INT, 0, MPI_COMM_WORLD );
    logfile = (char *) malloc( sizeof( char ) * str_len );
    MPI_Bcast( logfile, str_len, MPI_CHAR, 0, MPI_COMM_WORLD );
  }

  if (np==1) {
    fprintf( stderr, "Only one process started, cannot run.  The algorithm used in\n");
    fprintf( stderr, "this program requires at least 2 processes.\n\n");
    exit( 1 );
  }

  MPI_Bcast( &width,  1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &height, 1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &rmin, sizeof( NUM ), MPI_BYTE, 0, MPI_COMM_WORLD );
  MPI_Bcast( &rmax, sizeof( NUM ), MPI_BYTE, 0, MPI_COMM_WORLD );
  MPI_Bcast( &imin, sizeof( NUM ), MPI_BYTE, 0, MPI_COMM_WORLD );
  MPI_Bcast( &imax, sizeof( NUM ), MPI_BYTE, 0, MPI_COMM_WORLD );
  MPI_Bcast( &boundary,   1, MPI_DOUBLE, 0, MPI_COMM_WORLD );
  MPI_Bcast( &bw,         1, MPI_INT,    0, MPI_COMM_WORLD );
  MPI_Bcast( &maxiter,    1, MPI_INT,    0, MPI_COMM_WORLD );
  MPI_Bcast( &breakout,   1, MPI_INT,    0, MPI_COMM_WORLD );
  MPI_Bcast( &tasksize,   1, MPI_INT,    0, MPI_COMM_WORLD );
  MPI_Bcast( &randomize,  1, MPI_INT,    0, MPI_COMM_WORLD );
  MPI_Bcast( &numColors,  1, MPI_INT,    0, MPI_COMM_WORLD );

  MPE_OpenGraphics( &graph, MPI_COMM_WORLD, display, -1, -1, width, height, 0 );
  colorArray = (MPE_Color *) malloc( numColors * sizeof( MPE_Color ) );
  MPE_MakeColorArray( graph, numColors, colorArray );
  MPE_Update( graph );

#if LOG
  MPE_Startlog();
#endif

  if (myid==MASTER_PROC) {
    Master_Proc( width, height, maxiter, boundary,
                 rmin, rmax, imin, imax, tasksize, randomize );
  } else {								     /* sub-process */
  Slave_Proc( graph, width, height, maxiter, boundary,
	        rmin, rmax, imin, imax, bw, breakout,
	        tasksize, numColors );
  } /* not master process */

#if LOG
  if (myid==MASTER_PROC) {
    fprintf( stderr, "Writing logfile\n" );
  }
  MPE_Finishlog( logfile );
#endif
  MPI_Finalize();
  if (myid==MASTER_PROC) {
    fprintf( stderr, "Press <Return> to close window\n" );
  }
  if (!myid) {
    while (getchar()!='\n');
  }
}

