/*
 *  $Id: chsync.c,v 1.3 1994/04/12 16:02:57 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#ifndef lint
static char vc[] = "$Id: chsync.c,v 1.3 1994/04/12 16:02:57 gropp Exp $";
#endif

#include "mpid.h"

/*
   This file contains routines to keep track of synchronous send messages;
   they must be explicitly acknowledged.
 */

typedef struct _MPID_SyncId {
    int id;
    MPIR_SHANDLE *dmpi_send_handle;
    MPID_SHANDLE *mpid_send_handle;
    struct _MPID_SyncId *next;
    } MPID_SyncId;

static MPID_SyncId *head = 0;
static int         CurId = 0;
/* Get a message id for this message and add to the list */

void MPID_NX_Get_Sync_Id( mode, dmpi_handle, mpid_handle )
int *mode;
MPIR_SHANDLE *dmpi_handle;
MPID_SHANDLE *mpid_handle;
{
MPID_SyncId *new;

*mode		      = *mode | (CurId << 3);
new		      = (MPID_SyncId *)malloc(sizeof(MPID_SyncId));
if (!new) MPID_NX_Abort( MPI_ERR_EXHAUSTED );
new->id		      = CurId++;
new->dmpi_send_handle = dmpi_handle;
new->mpid_send_handle = mpid_handle;
new->next	      = head;
head		      = new;
}

int MPID_NX_Lookup_SyncAck( mode, dmpi_send_handle, mpid_send_handle )
int mode;
MPIR_SHANDLE **dmpi_send_handle;
MPID_SHANDLE **mpid_send_handle;
{
MPID_SyncId *cur, *last;
int         modeid = mode >> 3;
last = 0;
cur  = head;
*dmpi_send_handle = 0;
*mpid_send_handle = 0;
while (cur) {
    if (cur->id == modeid) {
	*dmpi_send_handle = cur->dmpi_send_handle;
	*mpid_send_handle = cur->mpid_send_handle;
	if (last) last->next = cur->next;
	else      head       = cur->next;
	free(cur );
	return;
	}
    last = cur;
    cur  = cur->next;
    }
}

/* Process a synchronization acknowledgement */
MPID_SyncAck( mode )
int mode;
{
MPIR_SHANDLE *dmpi_send_handle;
MPID_SHANDLE *mpid_send_handle;

/* This is an acknowledgement of a synchronous send; look it up and
   mark as completed */
MPID_NX_Lookup_SyncAck( mode, &dmpi_send_handle, &mpid_send_handle );
if (!dmpi_send_handle) {
    fprintf( stderr, "Error in processing sync ack!\n" );
    return MPI_SUCCESS;
    }
dmpi_send_handle->completed = 1;
mpid_send_handle->done	    = 1;
}

/* return an acknowledgment */
MPID_SyncReturnAck( mode, from )
int mode, from;
{
MPID_PACKET pkt;

pkt.mode = (mode & ~ 0x7) | MPIR_MODE_SYNC_ACK;
/* Rest of packet is irrelavent */
_csend(MPID_PT2PT_TAG,&pkt,MPID_HEADER_LEN,from,0);
}
