/*
 *  $Id: chrecv.c,v 1.16 1994/04/15 23:16:10 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */


#ifndef lint
static char vcid[] = "$Id: chrecv.c,v 1.16 1994/04/15 23:16:10 gropp Exp $";
#endif /* lint */

#include "mpid.h"
#ifndef MPI_ANY_TAG
#define MPI_ANY_SOURCE 	(-1)
#define MPI_ANY_TAG	(-1)
#endif
static int DebugFlag = 0;

void MPID_SetDebugFlag( f )
int f;
{
DebugFlag = f;
}
void MPID_SetRecvDebugFlag( f )
int f;
{
DebugFlag = f;
}

/* 
  This file contains the routines to handle receiving a message

  Because we don't know the length of messages (at least the long ones with
  tag MPID_PT2PT2_TAG(src) tags), we never post a receive.  Rather, we have
  a MPID_ch_check_incoming routine that looks for headers.  Note that
  messages sent from a source to destination with the MPID_PT2PT2_TAG(src)
  are ordered (we assume that the message-passing system preserves order).
 */

int MPID_CH_Process_unexpected( dmpi_recv_handle, dmpi_unexpected )
MPIR_RHANDLE *dmpi_recv_handle, *dmpi_unexpected;
{
MPID_RHANDLE *mpid_recv_handle;
MPID_RHANDLE *mpid_recv_handle_unex;
MPID_PACKET  pkt;	

mpid_recv_handle = &dmpi_recv_handle->dev_rhandle;

if (DebugFlag) {
    printf( "[%d] Found message in unexpected queue\n", PImytid );
    fflush( stdout );
    }
/* Copy relevant data to recv_handle */
mpid_recv_handle_unex      = &dmpi_unexpected->dev_rhandle;
dmpi_recv_handle->source   = dmpi_unexpected->source;
dmpi_recv_handle->tag      = dmpi_unexpected->tag;
dmpi_recv_handle->totallen = mpid_recv_handle_unex->bytes_as_contig;
dmpi_recv_handle->dev_rhandle.is_posted = 0;
if (DebugFlag) {
    printf( "[%d] Found message in temp area of %d bytes...\n", PImytid,
	    mpid_recv_handle_unex->bytes_as_contig );
    fflush( stdout );
    }
if (mpid_recv_handle->start) {
    memcpy( mpid_recv_handle->start, mpid_recv_handle_unex->temp,
	   mpid_recv_handle_unex->bytes_as_contig );
    }
else {
    if (mpid_recv_handle_unex->bytes_as_contig > 0) {
	DMPI_put_totallen( dmpi_recv_handle, 
			  mpid_recv_handle_unex->bytes_as_contig );
	DMPI_put_from_contig( dmpi_recv_handle, 
			     (void *)(mpid_recv_handle_unex->temp),
			     mpid_recv_handle_unex->bytes_as_contig );
	}
    }
FREE( mpid_recv_handle_unex->temp );

/* Return the synchronization message */
if ((mpid_recv_handle_unex->mode & 0x7) == MPIR_MODE_SYNCHRONOUS) {
    MPID_SyncReturnAck( mpid_recv_handle_unex->mode,dmpi_recv_handle->source );
    }

mpid_recv_handle->temp = 0;      /* In case of a cancel */
mpid_recv_handle->done = YES;
DMPI_mark_recv_completed(dmpi_recv_handle);

/* Recover dmpi_unexpected */
DMPI_free_unexpected( dmpi_unexpected );
}


/*
   Post a receive.

   Since the Chameleon implementation lets the underlying message transport
   layer handle this, there isn't much to do.  We DO handle the case of
   posting a BLOCKING receive by doing BLOCKING waits on incoming messages.

   Otherwise, we simply try to handle any receives that are ready for
   processing.
 */
int MPID_CH_post_recv( dmpi_recv_handle, is_available ) 
MPIR_RHANDLE *dmpi_recv_handle;
int          *is_available;
{
MPID_RHANDLE *mpid_recv_handle;
MPID_RHANDLE *mpid_recv_handle_unex;
MPIR_RHANDLE *dmpi_unexpected;
int          found;
MPID_PACKET  pkt;	

mpid_recv_handle = &dmpi_recv_handle->dev_rhandle;
*is_available    = 0;

if (DebugFlag) {
    printf( "[%d] starting recv for tag = %d, source = %d, ctx = %d\n", 
	    PImytid, dmpi_recv_handle->tag, dmpi_recv_handle->source,
	    dmpi_recv_handle->contextid );
    fflush( stdout );
    }
/* At this time, we check to see if the message has already been received */
DMPI_search_unexpected_queue( dmpi_recv_handle->source, 
		   dmpi_recv_handle->tag, dmpi_recv_handle->contextid, 
		   &found, 1, &dmpi_unexpected );
if (found) {
    if (DebugFlag) {
	printf( "[%d] found in unexpected queue\n", PImytid );
	fflush( stdout );
	}
    MPID_CH_Process_unexpected( dmpi_recv_handle, dmpi_unexpected );
    *is_available = 1;
    return MPI_SUCCESS;
    }

/* Add to the posted receive queue */
MPIR_enqueue( &MPIR_posted_recvs, dmpi_recv_handle, MPIR_QRHANDLE );
dmpi_recv_handle->dev_rhandle.is_posted = 1;

/* If we got here, the message is not yet available */
if (!mpid_recv_handle->is_non_blocking) {
    if (DebugFlag) {
	printf( "[%d] Blocking recv; starting wait loop\n", PImytid );
	fflush( stdout );
	}
    while (!dmpi_recv_handle->completed) {
	(void)MPID_CH_check_incoming( MPID_BLOCKING );
	}
    }

else {
    if (DebugFlag) {
	printf( "[%d] About to do a non-blocking check of incoming messages\n",
	        PImytid );
	fflush( stdout );
	}
    while (MPID_CH_check_incoming( MPID_NOTBLOCKING ) != -1) ;
    }

if (DebugFlag) {
    printf( "[%d] Exiting post receive\n", PImytid );
    fflush( stdout );
    }

return MPI_SUCCESS;
}

/* Check for incoming messages.
    Input Parameter:
.   is_blocking - true if this routine should block until a message is
    available

    Returns -1 if nonblocking and no messages pending
 */    
int MPID_CH_check_incoming( is_blocking )
MPID_BLOCKING_TYPE is_blocking;
{
MPID_PACKET  pkt;	
int          from;
char         *address;
MPIR_RHANDLE *dmpi_recv_handle;
int          actual_len;
MPID_RHANDLE *mpid_recv_handle;
int          is_posted;

/* If nonblocking and no headers available, exit */
#ifndef pvm3
if (is_blocking == MPID_NOTBLOCKING) {
    if (!PInprobe( MPID_PT2PT_TAG )) return -1;
    }
PIbrecv( MPID_PT2PT_TAG, &pkt, sizeof(MPID_PACKET), MSG_OTHER );
#else
/* pvm3 doesn't have a real probe, but what they do have meets the 
   semantics that we need here, though it is somewhat painful... 
   All this to save  the user a single routine call in the case where
   a probe is immediately followed by a recv.  Heaven help you if you
   use the probe to decide to call some other code to process the 
   message... 
*/
{
int bufid, bytes, msgtype; 
if (is_blocking == MPID_NOTBLOCKING) {
    if ((bufid = pvm_nrecv( -1, MPID_PT2PT_TAG )) <= 0) return -1;
    }
pvm_bufinfo( bufid, &bytes, &msgtype, &__PVMFROMTID );
pvm_upkint( (int *)&pkt, bytes / sizeof(int), 1 );
__PVMFROM = -1;
}
#endif

from = PIfrom();
/* We should check the size here for internal errors .... */

if ((pkt.mode & 0x7) == MPIR_MODE_SYNC_ACK) {
    MPID_SyncAck( pkt.mode );
    return MPI_SUCCESS;    
    }

if (DebugFlag) {
    printf( "[%d] received message for tag = %d, source = %d, ctx = %d\n", 
	    PImytid, pkt.tag, from, pkt.context_id );
    fflush( stdout );
    }
/* Is the message expected or not? */
DMPI_msg_arrived( pkt.lrank, pkt.tag, pkt.context_id, 
                  &dmpi_recv_handle, &is_posted );
mpid_recv_handle = &dmpi_recv_handle->dev_rhandle;
if (is_posted) {
    if (mpid_recv_handle->start) {
	dmpi_recv_handle->totallen = pkt.len;
	memcpy( mpid_recv_handle->start, (void *)(pkt.buffer), 
	        MPID_MIN(pkt.len,MPID_PACKET_SIZE) );
	}
    else {
	DMPI_put_totallen( dmpi_recv_handle, pkt.len );
	DMPI_put_from_contig( dmpi_recv_handle, (void *)(pkt.buffer),
			     MPID_MIN(pkt.len,MPID_PACKET_SIZE) );
	}
    if (DebugFlag) {
	printf( "[%d] Got data from packet\n", PImytid );
	fflush( stdout );
	}
    /* Handle long messages... */
    if (pkt.len > MPID_PACKET_SIZE) {
	if (DebugFlag) {
	    printf( "[%d] Receiving a long message from %d\n", PImytid, from );
	    fflush( stdout );
	    }
        /* Question: who is responsible for this allocation? */
	if (mpid_recv_handle->start) {
	    PIbrecv( MPID_PT2PT2_TAG(from), 
		    (((char *)mpid_recv_handle->start) + MPID_PACKET_SIZE), 
		    pkt.len-MPID_PACKET_SIZE,
		    MSG_OTHER );
	    }
	else {
	    DMPI_put_into_contig( dmpi_recv_handle, (void **)(&address), 
				 pkt.len - MPID_PACKET_SIZE, &actual_len );
	    PIbrecv( MPID_PT2PT2_TAG(from), address, pkt.len-MPID_PACKET_SIZE,
		    MSG_OTHER );
	    }
	}
    if ((pkt.mode & 0x7) == MPIR_MODE_SYNCHRONOUS) {
	MPID_SyncReturnAck( pkt.mode, from );
	}
    DMPI_mark_recv_completed(dmpi_recv_handle);
    mpid_recv_handle->done = YES;
    }
else {
    /* Note that DMPI_msg_arrived ALSO adds this message to the 
       unexpected queue */
    /* MPIR_enqueue( &unexpected_recvs, dmpi_recv_handle, MPIR_QRHANDLE ); */
    /* We buffer the message on the receiver.   There are, of course, other
       strategies */
    address = (char *)MALLOC( pkt.len ); 
    if (!address) return MPI_ERR_EXHAUSTED;
    memcpy( address, pkt.buffer, MPID_MIN(pkt.len,MPID_PACKET_SIZE) );
    if (pkt.len > MPID_PACKET_SIZE) {
	/* Message has "extension" part */
        PIbrecv( MPID_PT2PT2_TAG(from), address + MPID_PACKET_SIZE, 
		 pkt.len-MPID_PACKET_SIZE, MSG_OTHER );
	}
    mpid_recv_handle->temp            = address;
    mpid_recv_handle->bytes_as_contig = pkt.len;
    /* We also need to remember that the mode of this message incase
       it was a Synchronous message */
#ifdef DEBUG_READY
    if ((pkt.mode & 0x7) == MPIR_MODE_READY) {
	/* Issue error message-should never get an unexpected message 
	   in ready mode */
	(*MPID_ErrorHandler)( 1, 
			  "Received ready message without matching receive"  );
	}
#endif
    mpid_recv_handle->mode = pkt.mode;
    }
return MPI_SUCCESS;    
}


/*
    This routine completes a particular receive.  It does this by processing
    incoming messages until the indicated message is received.
 */
int MPID_CH_complete_recv( dmpi_recv_handle ) 
MPIR_RHANDLE *dmpi_recv_handle;
{
MPID_RHANDLE *mpid_recv_handle = &dmpi_recv_handle->dev_rhandle;

while (!mpid_recv_handle->done) {
    (void)MPID_CH_check_incoming( MPID_BLOCKING );
    }
return MPI_SUCCESS;
}


/*
   Special case code for blocking receive.  The "common" case is handled with
   straight-through code; uncommon cases call routines.
 */
int MPID_CH_blocking_recv( dmpi_recv_handle ) 
MPIR_RHANDLE *dmpi_recv_handle;
{
MPID_RHANDLE *mpid_recv_handle, *mpid_recv_handle_unex;
MPIR_RHANDLE *dmpi_unexpected, *dmpi_save_recv_handle;
MPID_PACKET  pkt;	
int          found, from, actual_len, is_posted, tag, source, context_id;
char         *address;

mpid_recv_handle = &dmpi_recv_handle->dev_rhandle;

if (DebugFlag) {
    printf( 
	"[%d] starting blocking recv for tag = %d, source = %d, ctx = %d\n", 
	    PImytid, dmpi_recv_handle->tag, dmpi_recv_handle->source,
	    dmpi_recv_handle->contextid );
    fflush( stdout );
    }
/* At this time, we check to see if the message has already been received */
DMPI_search_unexpected_queue( dmpi_recv_handle->source, 
		   dmpi_recv_handle->tag, dmpi_recv_handle->contextid, 
		   &found, 1, &dmpi_unexpected );
if (found) {
    MPID_CH_Process_unexpected( dmpi_recv_handle, dmpi_unexpected );
    return MPI_SUCCESS;
    }

dmpi_recv_handle->dev_rhandle.is_posted = 1;
dmpi_save_recv_handle = dmpi_recv_handle;
/* If we got here, the message is not yet available */
if (DebugFlag) {
    printf( "[%d] Blocking recv; starting wait loop\n", PImytid );
    fflush( stdout );
    }
tag	   = dmpi_recv_handle->tag;
context_id = dmpi_recv_handle->contextid;
source	   = dmpi_recv_handle->source;
while (!dmpi_save_recv_handle->completed) {
    PIbrecv( MPID_PT2PT_TAG, &pkt, sizeof(MPID_PACKET), MSG_OTHER );
    from = PIfrom();
    /* We should check the size here for internal errors .... */

    if ((pkt.mode & 0x7) == MPIR_MODE_SYNC_ACK) {
	MPID_SyncAck( pkt.mode );
	continue;
	}
    if (DebugFlag) {
	printf( "[%d] received message for tag = %d, source = %d, ctx = %d\n", 
	       PImytid, pkt.tag, from, pkt.context_id );
	fflush( stdout );
	}
    if (pkt.context_id == context_id && 
	(pkt.tag   == tag    || tag == MPI_ANY_TAG) &&
	(pkt.lrank == source || source == MPI_ANY_SOURCE)) {
	/* Found the message that I'm waiting for (it was never queued) */
	is_posted = 1;
	dmpi_recv_handle	 = dmpi_save_recv_handle;
	dmpi_recv_handle->tag	 = pkt.tag;
	dmpi_recv_handle->source = pkt.lrank;
	}
    else {
	/* Unexpected message */
	DMPI_msg_arrived( pkt.lrank, pkt.tag, pkt.context_id, 
			  &dmpi_recv_handle, &is_posted );
	}
    mpid_recv_handle = &dmpi_recv_handle->dev_rhandle;
    if (is_posted) {
	if (mpid_recv_handle->start) {
	    dmpi_recv_handle->totallen = pkt.len;
	    memcpy( mpid_recv_handle->start, (void *)(pkt.buffer), 
		   MPID_MIN(pkt.len,MPID_PACKET_SIZE) );
	    }
	else {
	    DMPI_put_totallen( dmpi_recv_handle, pkt.len );
	    DMPI_put_from_contig( dmpi_recv_handle, (void *)(pkt.buffer),
				 MPID_MIN(pkt.len,MPID_PACKET_SIZE) );
	    }
	if (DebugFlag) {
	    printf( "[%d] Got data from packet\n", PImytid );
	    fflush( stdout );
	    }
	/* Handle long messages... */
	if (pkt.len > MPID_PACKET_SIZE) {
	    /* MPID_ProcessLongMessage( from, mpid_recv_handle ); */
	    if (DebugFlag) {
		printf( "[%d] Receiving a long message from %d\n", 
		        PImytid, from );
		fflush( stdout );
		}
	    /* Question: who is responsible for this allocation? */
	    if (mpid_recv_handle->start) {
		PIbrecv( MPID_PT2PT2_TAG(from), 
		       (((char *)mpid_recv_handle->start) + MPID_PACKET_SIZE), 
			pkt.len-MPID_PACKET_SIZE,
			MSG_OTHER );
		}
	    else {
		DMPI_put_into_contig( dmpi_recv_handle, (void **)(&address), 
				     pkt.len - MPID_PACKET_SIZE, &actual_len );
		PIbrecv( MPID_PT2PT2_TAG(from), address, 
			 pkt.len-MPID_PACKET_SIZE, MSG_OTHER );
		}
	    }
	if ((pkt.mode & 0x7) == MPIR_MODE_SYNCHRONOUS) {
	    MPID_SyncReturnAck( pkt.mode, from );
	    }
	DMPI_mark_recv_completed(dmpi_recv_handle);
	mpid_recv_handle->done = YES;
	}
    else {
	/* We buffer the message on the receiver.   There are, of course, other
	   strategies */
	address = (char *)MALLOC( pkt.len ); 
	if (!address) return MPI_ERR_EXHAUSTED;
	memcpy( address, pkt.buffer, MPID_MIN(pkt.len,MPID_PACKET_SIZE) );
	if (pkt.len > MPID_PACKET_SIZE) {
	    /* Message has "extension" part */
	    PIbrecv( MPID_PT2PT2_TAG(from), address + MPID_PACKET_SIZE, 
		    pkt.len-MPID_PACKET_SIZE, MSG_OTHER );
	    }
	mpid_recv_handle->temp            = address;
	mpid_recv_handle->bytes_as_contig = pkt.len;
	/* We also need to remember that the mode of this message incase
	   it was a Synchronous message */
#ifdef DEBUG_READY
	if ((pkt.mode & 0x7) == MPIR_MODE_READY) {
	    /* Issue error message-should never get an unexpected message 
	       in ready mode */
	    }
#endif
	mpid_recv_handle->mode = pkt.mode;
	}
    }
}

