/*
 *  $Id: alltoall.c,v 1.13 1994/04/12 20:17:49 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#ifndef lint
static char vcid[] = "$Id: alltoall.c,v 1.13 1994/04/12 20:17:49 doss Exp $";
#endif /* lint */

#include <mpi.h>
#include "coll.h"

/*@

MPI_Alltoall - Send data from all to all processes

Input Parameters:
. sendbuf - starting address of send buffer (choice) 
. sendcount - number of elements in send buffer (integer) 
. sendtype - data type of send buffer elements (handle) 
. recvcount - number of elements received from any process (integer) 
. recvtype - data type of receive buffer elements (handle) 
. comm - communicator (handle) 

Output Parameter:
. recvbuf - address of receive buffer (choice) 
@*/
int MPI_Alltoall( sendbuf, sendcount, sendtype, 
                  recvbuf, recvcnt, recvtype, comm )
void             *sendbuf;
int               sendcount;
MPI_Datatype      sendtype;
void             *recvbuf;
int               recvcnt;
MPI_Datatype      recvtype;
MPI_Comm          comm;
{
  int        size, rank, i;
  MPI_Aint   extent;
  int        errno = MPI_SUCCESS;
  MPI_Status status;
  int        flag;

  /* Check for invalid arguments */
  if ( (comm == MPI_COMM_NULL) )
	return MPIR_ERROR(comm, MPI_ERR_COMM,
					  "NULL communicator in MPI_ALLTOALL" ); 

  /* Check for intra-communicator */
  MPI_Comm_test_inter ( comm, &flag );
  if (flag) 
    return MPIR_ERROR(comm, MPI_ERR_COMM,
					  "Inter-communicator invalid in MPI_ALLTOALL");
  
  /* Get size and rank and switch to collective communicator */
  MPI_Comm_size ( comm, &size );
  MPI_Comm_rank ( comm, &rank );
  comm = comm->comm_coll;

  /* Send to all - receive from all */
  MPI_Type_extent ( recvtype, &extent );
  for ( i=0; i<size; i++ ) { 
       MPI_Sendrecv( (void *)((int)sendbuf+i*sendcount*extent), 
		    sendcount, sendtype, i, MPIR_ALLTOALL_TAG,
                     (void *)((int)recvbuf+i*recvcnt*extent), 
		    recvcnt, recvtype, i, MPIR_ALLTOALL_TAG,
		     comm, &status );
  }

  return (errno);
}
