/*
 *  $Id: scatter.c,v 1.14 1994/04/13 21:57:08 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */


#ifndef lint
static char vcid[] = "$Id: scatter.c,v 1.14 1994/04/13 21:57:08 doss Exp $";
#endif /* lint */

#include <mpi.h>
#include "coll.h"

/*@

MPI_Scatter - Send data from one task to all other tasks in a group

Input Parameters:
. sendbuf - address of send buffer (choice, significant 
only at root) 
. sendcount - number of elements sent to each process 
(integer, significant only at root) 
. sendtype - data type of send buffer elements 
(significant only at root) (handle) 
Input Parameters:
. recvcount - number of elements in receive buffer (integer) 
. recvtype - data type of receive buffer elements (handle) 
. root - rank of sending process (integer) 
. comm - communicator (handle) 

Output Parameter:
. recvbuf - address of receive buffer (choice) 

@*/
int MPI_Scatter ( sendbuf, sendcnt, sendtype, 
		  recvbuf, recvcnt, recvtype, 
		  root, comm )
void             *sendbuf;
int               sendcnt;
MPI_Datatype      sendtype;
void             *recvbuf;
int               recvcnt;
MPI_Datatype      recvtype;
int               root;
MPI_Comm          comm;
{
  MPI_Status status;
  MPI_Aint   extent;
  int        rank, size, i;
  int        errno = MPI_SUCCESS;
  int        flag;
  
  /* Get size and rank */
  MPI_Comm_size ( comm, &size );
  MPI_Comm_rank ( comm, &rank );

  /* Check for invalid arguments */
  if ( ( (comm            == MPI_COMM_NULL)  && (errno = MPI_ERR_COMM) )   ||
       ( (root            <  0)              && (errno = MPI_ERR_ROOT) )   || 
	   ( (root            >= size)           && (errno = MPI_ERR_ROOT) )   || 
       ( ((recvcnt>0)&&(recvbuf==(void *)0)) && (errno = MPI_ERR_BUFFER) ) )
      return MPIR_ERROR( comm, errno, "Error in MPI_SCATTER" ); 
  
  /* Check for intra-communicator */
  MPI_Comm_test_inter ( comm, &flag );
  if (flag) 
    return MPIR_ERROR(comm, MPI_ERR_COMM,
                      "Inter-communicator invalid in MPI_SCATTER");

  /* Switch communicators to the hidden collective */
  comm = comm->comm_coll;

  /* Get the size of the send type */
  MPI_Type_extent ( sendtype, &extent );

  /* If I'm the root, send messages to the rest of 'em */
  if ( rank == root ) {
    for ( i=0; i<root; i++ )
      MPI_Send( (void *)((int)sendbuf+i*sendcnt*extent), sendcnt, 
		sendtype, i, MPIR_SCATTER_TAG, comm);

    MPI_Sendrecv ( (void *)((int)sendbuf+rank*sendcnt*extent),sendcnt,sendtype, rank, MPIR_SCATTER_TAG,
		   recvbuf, recvcnt, recvtype, rank, MPIR_SCATTER_TAG, 
		   comm, &status );

    for ( i=root+1; i<size; i++ )
      MPI_Send( (void *)((int)sendbuf+i*sendcnt*extent), sendcnt, 
		sendtype, i, MPIR_SCATTER_TAG, comm);
  }
  else 
    MPI_Recv(recvbuf,recvcnt,recvtype,root,MPIR_SCATTER_TAG,comm,&status);

  return (errno);
}
