/*
 *  $Id: attr_putval.c,v 1.10 1994/04/07 08:55:54 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include <mpi.h>

#ifndef MPIR_TRUE
#define MPIR_TRUE  1
#define MPIR_FALSE 0
#endif

/*@

MPI_Attr_put - stores attribute value associated with a key

Input Parameters:
. comm - communicator to which attribute will be attached (handle) 
. keyval - key value, as returned by  MPI_KEYVAL_CREATE (integer) 
. attribute_val - attribute value 

@*/
int MPI_Attr_put ( comm, keyval, attr_value )
MPI_Comm comm;
int      keyval;
void     *attr_value;
{
  MPIR_HBT_node *attr;
  MPIR_Attr_key *attr_key = (MPIR_Attr_key *)keyval;

  /* Check for valid arguments */
  if ( comm == MPI_COMM_NULL )
	return MPIR_ERROR( comm, MPI_ERR_COMM, "Error in MPI_Attr_put" );

  MPIR_HBT_lookup(comm->attr_cache, keyval, &attr);
  if (attr == (MPIR_HBT_node *)0) {
	(void) MPIR_HBT_new_node ( keyval, attr_value, &attr );
	(void) MPIR_HBT_insert ( comm->attr_cache, attr );
  }
  else {
	if ( attr_key->delete_fn != (int(*)())0 )
	  (void) attr_key->delete_fn( comm, keyval, attr_key->extra_state );
	attr->value = attr_value;
  }
  return (MPI_SUCCESS);
}
