/*
 *  $Id: comm_free.c,v 1.13 1994/04/13 21:43:45 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include <mpi.h>
#include "mpisys.h"

/*@

MPI_Comm_free - Marks the communicator object for deallocation

Input Parameter:
. comm - communicator to be destroyed (handle) 

@*/
int MPI_Comm_free ( comm )
MPI_Comm *comm;
{
  int errno = MPI_SUCCESS;

  /* Check for bad arguments */
  if ( ( (comm    == (MPI_Comm *)0) && (errno = MPI_ERR_ARG) ) ||
	   ( ((*comm) == MPI_COMM_NULL) && (errno = MPI_ERR_COMM) ) )
	return MPIR_ERROR( MPI_COMM_NULL, errno,
					  "Error in MPI_COMM_FREE" );

  /* Free group */
  if ( (*comm)->ref_count <= 1 ) {

	/* Free the contexts used by this communicator */
	if ((*comm)->comm_type == MPIR_INTRA)
	  (void) MPIR_Context_dealloc ( (*comm), 2, (*comm)->recv_context );
	else
	  (void) MPIR_Context_dealloc ( (*comm), 3, (*comm)->recv_context );
	
	/* Free cache information */
	MPIR_Attr_free_tree ( *comm );
	
	/* Free collective communicator(s) and group */
	MPI_Group_free ( &((*comm)->group) );
	if ((*comm)->comm_type == MPIR_INTER)
	  FREE( (*comm)->comm_coll->comm_coll );
	FREE( (*comm)->comm_coll );
	
	/* Free topology */
	if ( (*comm)->topology.type == MPI_CART ) 
	  FREE( (*comm)->topology.tag.cart.dims );
	else if ( (*comm)->topology.type == MPI_GRAPH )
	  FREE( (*comm)->topology.tag.graph.index );
	
	/* Free comm structure */
	FREE( *comm );
  }
  else 
	(*comm)->ref_count--;

  /* Set comm to null */
  (*comm) = MPI_COMM_NULL;
	
  return (errno);
}
