/*
 *  $Id: init.c,v 1.23 1994/04/18 13:39:57 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */


#ifndef lint
static char vcid[] = "$Id: init.c,v 1.23 1994/04/18 13:39:57 gropp Exp $";
#endif /* lint */
#include "mpi.h"
#include "mpisys.h"

#ifdef MPI_rs6000
#define mpir_init_fdtes_ mpir_init_fdtes
#define mpir_init_fcm_   mpir_init_fcm
#define mpir_init_fop_   mpir_init_fop
#endif

#ifdef MPI_cray
#define mpir_init_fdtes_ MPIR_INIT_FDTES
#define mpir_init_fcm_   MPIR_INIT_FCM
#define mpir_init_fop_   MPIR_INIT_FOP
#endif

/* Global memory management variables for fixed-size blocks */
void *MPIR_shandles;        /* sbcnst MPIR_SHANDLES */
void *MPIR_rhandles;        /* sbcnst MPIR_RHANDLES */
void *MPIR_dtes;      /* sbcnst datatype elements */
void *MPIR_qels;      /* sbcnst queue elements */
void *MPIR_fdtels; /* sbcnst flat datatype elements */
void *MPIR_hbts;   /* sbcnst height balanced tree roots for cacheing */
void *MPIR_hbt_els;/* sbcnst height balanced tree nodes for cacheing */

/* Global queues */
MPIR_QHDR MPIR_posted_recvs;
MPIR_QHDR MPIR_unexpected_recvs;

/* Global pre-assigned datatypes */
MPIR_DATATYPE *MPIR_int_dte;
MPIR_DATATYPE *MPIR_float_dte;
MPIR_DATATYPE *MPIR_double_dte;
MPIR_DATATYPE *MPIR_long_dte;
MPIR_DATATYPE *MPIR_complex_dte; 
MPIR_DATATYPE *MPIR_dcomplex_dte;
MPIR_DATATYPE *MPIR_short_dte;
MPIR_DATATYPE *MPIR_char_dte;
MPIR_DATATYPE *MPIR_byte_dte;
MPIR_DATATYPE *MPIR_uchar_dte;
MPIR_DATATYPE *MPIR_ushort_dte;
MPIR_DATATYPE *MPIR_ulong_dte;
MPIR_DATATYPE *MPIR_uint_dte;
MPIR_DATATYPE *MPIR_longdouble_dte;
MPIR_DATATYPE *MPIR_packed_dte;
MPIR_DATATYPE *MPIR_ub_dte;
MPIR_DATATYPE *MPIR_lb_dte;
MPIR_DATATYPE *MPIR_Init_basic_datatype( );

/* Global MPIR process id (from device) */
int MPIR_tid;

/* Predefined combination functions */
MPI_Op MPI_MAX, MPI_MIN, MPI_SUM, MPI_PROD, MPI_LAND, MPI_BAND, 
               MPI_LOR, MPI_BOR, MPI_LXOR, MPI_BXOR, MPI_MAXLOC, MPI_MINLOC;

/* Permanent attributes */
int MPI_TAG_UB, MPI_HOST, MPI_IO;

/* Global buffer */
int MPIR_Buffer_size = 0;
void *MPIR_Buffer = 0;

/* Command-line flags */
int MPIR_Print_queues = 0;

/* Fortran logical values */
/* NOTE: we really have to get these for each system (possibly by calling a
    routine) and make sure that the MPI routines whose Fortran bindings
    use LOGICAL convert to/from these values 
 */
int MPIR_F_TRUE = 1, MPIR_F_FALSE = 0;

/* 
   If I want to use __STDC__, I need to include the full prototypes of
   these functions ... 
 */
extern void MPIR_MAXF();
extern void MPIR_MINF();
extern void MPIR_SUM();
extern void MPIR_PROD();
extern void MPIR_LAND();
extern void MPIR_BAND();
extern void MPIR_LOR();
extern void MPIR_BOR();
extern void MPIR_LXOR();
extern void MPIR_BXOR();
extern void MPIR_MAXLOC();
extern void MPIR_MINLOC();

/*@
   MPI_Init - Initialize the MPI execution environment

   Input Parameters:
.  argc - Pointer to the number of arguments 
.  argv - Pointer to the argument vector

   Command line arguments:
   MPI specifies no command-line arguments but does allow an MPI 
   implementation to make use of them.

.   -mpiqueue - print out the state of the message queues when MPI_FINALIZE
   is called.  All processors print; the output may be hard to decipher.  This
   is intended as a debugging aid.

.  -chdebug - (Chameleon device only) Print out the Chameleon device operations
.  -chmemdebug - (Chameleon device only) Print out a list of unreclaimed
   memory.  This requires that MPI be built with the -DMPIR_DEBUG_MEM
   switch.  This is intended for debugging the MPI implementation itself.
   
   Notes:
   Note that the Fortran binding for this routine has only the error return
   argument
@*/
int MPI_Init(argc,argv)
int  *argc;
char ***argv;
{
    int size;
    char **Argv;

    if (MPIR_Has_been_initialized) 
    return 
        MPIR_ERROR( (MPI_Comm)0, MPI_ERR_INIT, "Can not MPI_INIT again" );

    Argv = (char **)0;
    if (argv) Argv = *argv;
    MPID_INIT( argc, Argv );

    /* initialize memory allocation data structures */
    MPIR_shandles   = SBinit( sizeof( MPIR_SHANDLE ), 100, 100 );
    MPIR_rhandles   = SBinit( sizeof( MPIR_RHANDLE ), 100, 100 );
    MPIR_dtes       = SBinit( sizeof( MPIR_DATATYPE ), 100, 100 );
    MPIR_qels       = SBinit( sizeof( MPIR_QEL ), 100, 100 );
    MPIR_fdtels     = SBinit( sizeof( MPIR_FDTEL ), 100, 100 );
    MPIR_hbts       = SBinit( sizeof( MPIR_HBT ), 5, 5 );
    MPIR_hbt_els    = SBinit( sizeof( MPIR_HBT_node ), 20, 20);

    /* set up pre-defined data types */
    MPIR_int_dte      = MPIR_Init_basic_datatype( MPIR_INT, sizeof(int) );
    MPIR_float_dte    = MPIR_Init_basic_datatype( MPIR_FLOAT, sizeof(float) );
    MPIR_double_dte   = MPIR_Init_basic_datatype( MPIR_DOUBLE, 
					        sizeof( double ) );
    MPIR_complex_dte  = MPIR_Init_basic_datatype( MPIR_COMPLEX, 
						   2 * sizeof( float ) );
    MPIR_dcomplex_dte = MPIR_Init_basic_datatype( MPIR_DOUBLE_COMPLEX, 
						  2 * sizeof( double ) );
    MPIR_long_dte     = MPIR_Init_basic_datatype( MPIR_LONG, sizeof( long ) );
    MPIR_short_dte    = MPIR_Init_basic_datatype( MPIR_SHORT, 
						  sizeof( short ) );
    MPIR_char_dte     = MPIR_Init_basic_datatype( MPIR_CHAR, sizeof( char ) );
    MPIR_byte_dte     = MPIR_Init_basic_datatype( MPIR_BYTE, sizeof( char ) );
    MPIR_uchar_dte    = MPIR_Init_basic_datatype( MPIR_UCHAR, 
					       sizeof( unsigned char ) );
    MPIR_ushort_dte   = MPIR_Init_basic_datatype( MPIR_USHORT, 
					        sizeof( unsigned short ) );
    MPIR_ulong_dte    = MPIR_Init_basic_datatype( MPIR_ULONG, 
					       sizeof( unsigned long ) );
    MPIR_uint_dte     = MPIR_Init_basic_datatype( MPIR_UINT, 
					       sizeof( unsigned int ) );
    MPIR_packed_dte   = MPIR_Init_basic_datatype( MPIR_PACKED, 1 );
    MPIR_ub_dte	      = MPIR_Init_basic_datatype( MPIR_UB, 0 );
    MPIR_ub_dte->elements = 0;
    MPIR_ub_dte->count	  = 0;

    MPIR_lb_dte       = MPIR_Init_basic_datatype( MPIR_LB, 0 );
    MPIR_lb_dte->elements = 0;
    MPIR_lb_dte->count	  = 0;

/* rs6000's assert STDC but do not implement long double. */
#if defined(__STDC__) && !defined(MPI_rs6000)
    MPIR_longdouble_dte	= MPIR_Init_basic_datatype( MPIR_LONGDOUBLE, 
						    sizeof ( long double ) );
#endif

    /* Set the values of the Fortran versions */
    /* Logical and character aren't portable in the code below */
    mpir_init_fdtes_( &MPIR_int_dte, &MPIR_float_dte, &MPIR_double_dte,
                     &MPIR_complex_dte, &MPIR_dcomplex_dte,
                     &MPIR_int_dte, &MPIR_char_dte, 
                     &MPIR_byte_dte );

    /* initialize queues */
    MPIR_posted_recvs.first        = MPIR_posted_recvs.last        = NULL;
    MPIR_posted_recvs.maxlen       = MPIR_posted_recvs.currlen     = 0; 

    MPIR_unexpected_recvs.first    = MPIR_unexpected_recvs.last    = NULL;
    MPIR_unexpected_recvs.maxlen   = MPIR_unexpected_recvs.currlen = 0;

    /* GROUP_EMPTY is a valid empty group */
    MPI_GROUP_EMPTY     = MPIR_CreateGroup(0);

    MPI_COMM_WORLD      = NEW(MPIR_COMMUNICATOR);    
    MPI_COMM_WORLD->comm_type   = MPIR_INTRA;
    MPID_mysize( &size );
    MPID_myrank( &MPIR_tid);
    MPI_COMM_WORLD->group   = MPIR_CreateGroup( size );
    MPI_COMM_WORLD->local_group = MPI_COMM_WORLD->group;
    MPIR_SetToIdentity( MPI_COMM_WORLD->group );

    MPI_COMM_WORLD->send_context = MPIR_WORLD_PT2PT_CONTEXT;
    MPI_COMM_WORLD->recv_context = MPIR_WORLD_PT2PT_CONTEXT;

    MPI_COMM_WORLD->topology.type = MPI_UNDEFINED;

    MPI_COMM_WORLD->error_handler = MPI_ERRORS_ARE_FATAL;

    MPIR_Comm_make_coll ( MPI_COMM_WORLD );

    MPIR_HBT_new_tree ( &(MPI_COMM_WORLD->attr_cache) );

    MPI_COMM_WORLD->comm_cache    = 0;

    /* Predefined attributes for MPI_COMM_WORLD */
    MPI_Keyval_create( (int (*)())0, (int (*)())0, &MPI_TAG_UB, (void *)0 );
    MPI_Keyval_create( (int (*)())0, (int (*)())0, &MPI_HOST,   (void *)0 );
    MPI_Keyval_create( (int (*)())0, (int (*)())0, &MPI_IO,     (void *)0 );
    ((MPIR_Attr_key *)(MPI_TAG_UB))->permanent = 1;
    ((MPIR_Attr_key *)(MPI_HOST))->permanent   = 1;
    ((MPIR_Attr_key *)(MPI_IO))->permanent     = 1;
    MPI_Attr_put( MPI_COMM_WORLD, MPI_TAG_UB, (void*)((1<<30)-1) );
    MPI_Attr_put( MPI_COMM_WORLD, MPI_HOST, (void*)MPI_UNDEFINED );
    /* The following isn't strictly correct, but I'm going to leave it
       in for now */
    MPI_Attr_put( MPI_COMM_WORLD, MPI_IO, (void*)MPI_ANY_SOURCE );

    /* COMM_SELF is the communicator consisting only of myself */
    MPI_COMM_SELF        = NEW(MPIR_COMMUNICATOR);    
    MPI_COMM_SELF->comm_type     = MPIR_INTRA;
    MPID_mysize( &size );
    MPID_myrank( &MPIR_tid);
    MPI_COMM_SELF->group     = MPIR_CreateGroup( 1 );
    MPI_COMM_SELF->local_group   = MPI_COMM_SELF->group;
    MPI_COMM_SELF->group->local_rank = 0;
    MPI_COMM_SELF->group->lrank_to_grank[0] = MPIR_tid;
    MPI_COMM_SELF->send_context  = MPIR_SELF_PT2PT_CONTEXT;
    MPI_COMM_SELF->recv_context  = MPIR_SELF_PT2PT_CONTEXT;
    MPI_COMM_SELF->topology.type = MPI_UNDEFINED;

    MPI_COMM_SELF->error_handler = MPI_ERRORS_ARE_FATAL;

    MPIR_Comm_make_coll ( MPI_COMM_SELF );

    MPIR_HBT_new_tree ( &(MPI_COMM_SELF->attr_cache) );

    MPI_COMM_SELF->comm_cache    = 0;

    /* Predefined combination functions */
    MPI_Op_create( MPIR_MAXF,   1, &MPI_MAX );
    MPI_Op_create( MPIR_MINF,   1, &MPI_MIN );
    MPI_Op_create( MPIR_SUM,    1, &MPI_SUM );
    MPI_Op_create( MPIR_PROD,   1, &MPI_PROD );
    MPI_Op_create( MPIR_LAND,   1, &MPI_LAND );
    MPI_Op_create( MPIR_BAND,   1, &MPI_BAND );
    MPI_Op_create( MPIR_LOR,    1, &MPI_LOR );
    MPI_Op_create( MPIR_BOR,    1, &MPI_BOR );
    MPI_Op_create( MPIR_LXOR,   1, &MPI_LXOR );
    MPI_Op_create( MPIR_BXOR,   1, &MPI_BXOR );
#ifdef FOO
    MPI_Op_create( MPIR_MAXLOC, 1, &MPI_MAXLOC );
    MPI_Op_create( MPIR_MINLOC, 1, &MPI_MINLOC );
    mpir_init_fop_( &MPI_MAX, &MPI_MIN, &MPI_SUM, &MPI_PROD, 
                    &MPI_LAND, &MPI_BAND,
                    &MPI_LOR, &MPI_BOR, &MPI_LXOR, &MPI_BXOR, 
                    &MPI_MAXLOC, &MPI_MINLOC );
#else
    { int d = 0;
    mpir_init_fop_( &MPI_MAX, &MPI_MIN, &MPI_SUM, &MPI_PROD, 
                    &MPI_LAND, &MPI_BAND,
                    &MPI_LOR, &MPI_BOR, &MPI_LXOR, &MPI_BXOR, 
                    &d, &d );
      }
#endif

    mpir_init_fcm_( &MPI_COMM_WORLD, &MPI_COMM_SELF, &MPI_GROUP_EMPTY );

    /* Search for "-mpi debug" options etc.  We need a better interface.... */
    if (argv && *argv) {
    int i;
    for (i=1; i<*argc; i++) {
        if ((*argv)[i] && strcmp( (*argv)[i], "-mpiqueue" ) == 0) {
        MPIR_Print_queues = 1;
        /* Should remove argument ... */
        }
        }
    }
#ifdef DEVICE_CHAMELEON
    if (argv && *argv && SYArgHasName( argc, *argv, 1, "-chdebug" )) {
    MPID_SetSendDebugFlag( 1 );
    MPID_SetRecvDebugFlag( 1 );
    }
    if (argv && *argv && SYArgHasName( argc, *argv, 1, "-chmemdebug" )) {
    MPID_SetSpaceDebugFlag( 1 );
    }
#endif
    /* barrier */
    MPIR_Has_been_initialized = 1;

    return MPI_SUCCESS;
}

MPIR_DATATYPE *MPIR_Init_basic_datatype( type, size )
int type, size;
{
MPIR_DATATYPE *new;

    new                 = (MPIR_DATATYPE *) SBalloc( MPIR_dtes );
    new->dte_type       = MPIR_FLOAT;
    new->committed      = YES;
    new->is_contig      = YES;
    new->lb             = 0;
    new->ub             = size;
    new->extent         = size;
    new->size           = size;
    new->align          = size;
    new->stride         = size;
    new->elements       = 1;
    new->count          = 1;
    new->blocklen       = 1;
    new->basic          = YES;
    new->old_type       = new;

return new;
}
