/*
 *  $Id: mperror.c,v 1.6 1994/04/01 23:33:03 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include "mpi.h"

/* 
   This calls the user-specified error handler.  If that handler returns,
   we return the error code 
 */
int MPIR_Error( comm, code, string, file, line )
MPI_Comm comm;
int              code, line;
char             *string, *file;
{
    int myid;
    char buf[MPI_MAX_ERROR_STRING];

    MPI_Error_string( code, MPI_MAX_ERROR_STRING, buf );
    if (!comm) 
	comm = MPI_COMM_WORLD;
    if (!comm || !comm->error_handler) {
	/* Errors are fatal is the default mode */
	myid = 0;
	fprintf( stderr, "%d - %s:%s\n", myid, 
		string ? string : "<NO ERROR MESSAGE>", buf );
	MPI_Abort( comm, code );
    }
    /* Call user routine */
    (comm->error_handler)( comm, &code, string, file, line );
    return code;
}

#ifdef __STDC__
void MPI_ERRORS_ARE_FATAL(  MPI_Comm comm, int *code, ... )
{
    char buf[MPI_MAX_ERROR_STRING];
    int myid = 0; 
    
    MPI_Error_string( *code, MPI_MAX_ERROR_STRING, buf );
/*
    fprintf( stderr, "%d - %s:%s\n", myid, 
	    string ? string : "<NO ERROR MESSAGE>", buf );
 */
    MPI_Abort( comm, *code );
}
#else
void MPI_ERRORS_ARE_FATAL(  comm, code, string, file, line )
MPI_Comm comm;
int      *code, line;
char     *string, *file;
{
    char buf[MPI_MAX_ERROR_STRING];
    int myid = 0; 
    
    MPI_Error_string( *code, MPI_MAX_ERROR_STRING, buf );
    fprintf( stderr, "%d - %s:%s\n", myid, 
	    string ? string : "<NO ERROR MESSAGE>", buf );
    MPI_Abort( comm, *code );
}
#endif

#ifdef __STDC__
void MPI_ERRORS_RETURN(  MPI_Comm comm, int *code, ... )
{
}
#else
void MPI_ERRORS_RETURN(  comm, code, string, file, line )
MPI_Comm comm;
int      *code, line;
char     *string, *file;
{
}
#endif

