/*
 *  $Id: ssend_init.c,v 1.5 1994/04/13 21:44:53 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */


#include "mpi.h"
#include "mpisys.h"

/*@
    MPI_Ssend_init - Build a handle for a synchronous send

Input Parameters:
. buf - initial address of send buffer (choice) 
. count - number of elements sent (integer) 
. datatype - type of each element (handle) 
. dest - rank of destination (integer) 
. tag - message tag (integer) 
. comm - communicator (handle) 
Output Parameter:
. request - communication request (handle) 
@*/
int MPI_Ssend_init( buf, count, datatype, dest, tag, comm, request )
void          *buf;
int           count;
MPI_Datatype  datatype;
int           dest;
int           tag;
MPI_Comm      comm;
MPI_Request   *request;
{

    MPIR_HANDLE *handleptr;

    *request                        = (MPI_Request) SBalloc( MPIR_shandles );
    handleptr                       = *request;
    memset( (char *)handleptr, 0, sizeof(MPIR_SHANDLE) );
    handleptr->type                 = MPIR_SEND;
    if (dest == MPI_PROC_NULL)
	handleptr->shandle.dest     = dest;
    else
	handleptr->shandle.dest     = comm->group->lrank_to_grank[dest];
    handleptr->shandle.tag          = tag;
    handleptr->shandle.contextid    = comm->send_context;
    handleptr->shandle.lrank        = 
	comm->local_group->lrank_to_grank[comm->local_group->local_rank];
    handleptr->shandle.mode         = MPIR_MODE_SYNCHRONOUS;
    handleptr->shandle.datatype     = datatype;
    handleptr->shandle.bufadd       = buf;
    handleptr->shandle.count        = count;
    handleptr->shandle.completed    = NO;
    MPID_alloc_send_handle(&((handleptr)->shandle.dev_shandle));
    MPID_set_send_is_nonblocking( &((handleptr)->shandle.dev_shandle), 1 );

    return MPI_SUCCESS;
}
