/*
 *  $Id: type_contig.c,v 1.1 1994/04/18 15:45:21 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include "mpi.h"
#include "mpisys.h"

#ifndef MPIR_TRUE
#define MPIR_TRUE  1
#define MPIR_FALSE 0
#endif


/*@
    MPI_Type_contiguous - Create a contiguous datatype
@*/
int MPI_Type_contiguous( count, old_type, newtype )
int          count;
MPI_Datatype old_type;
MPI_Datatype *newtype;
{
  MPIR_DATATYPE  *dteptr;
  int errno = MPI_SUCCESS;

  /* Check for bad arguments */
  if ( ( (old_type == MPI_DATATYPE_NULL) && (errno = MPI_ERR_TYPE) )  ||
	   ( (old_type == (*newtype))        && (errno = MPI_ERR_TYPE) )  ||
	   ( (count   <= 0)                  && (errno = MPI_ERR_COUNT) ) )
	return MPIR_ERROR( MPI_COMM_NULL, errno,
					  "Error in MPI_TYPE_CONTIGUOUS" );
	
  /* Create and fill in the datatype */
  dteptr = (*newtype) = (MPI_Datatype) SBalloc( MPIR_dtes );
  dteptr->dte_type    = MPIR_CONTIG;
  dteptr->committed   = MPIR_FALSE;
  dteptr->basic       = MPIR_FALSE;
  dteptr->ref_count   = 1;
  dteptr->align       = old_type->align;
  dteptr->stride      = 1;
  dteptr->blocklen    = 1;
  dteptr->is_contig   = old_type->is_contig;
  dteptr->elements    = count * old_type->elements;

  /* Take care of contiguous vs non-contiguous case */
  if (old_type->is_contig) {
	dteptr->old_type  = (MPI_Datatype)MPIR_Type_dup (old_type->old_type);
	dteptr->count     = count * old_type->count;
  }
  else {
	dteptr->old_type  = (MPI_Datatype)MPIR_Type_dup (old_type);
	dteptr->count     = count;
  }

  /* Set the upper/lower bounds and the extent and size */
  dteptr->lb          = dteptr->old_type->lb;
  dteptr->extent      = dteptr->count * dteptr->old_type->extent;
  dteptr->ub          = dteptr->lb + dteptr->extent;
  dteptr->size        = dteptr->count * dteptr->old_type->size;
  
  return (errno);
}
