/*
 *  $Id: type_hvec.c,v 1.1 1994/04/18 15:47:07 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include "mpi.h"
#include "mpisys.h"

#ifndef MPIR_TRUE
#define MPIR_TRUE  1
#define MPIR_FALSE 0
#endif


/*@
    MPI_Type_hvector - Create a vector (strided) datatype with offset in bytes
@*/
int MPI_Type_hvector( count, blocklen, stride, old_type, newtype )
int          count;
int 	     blocklen;
MPI_Aint     stride;
MPI_Datatype old_type;
MPI_Datatype *newtype;
{
  MPIR_DATATYPE  *dteptr;
  int errno = MPI_SUCCESS;

  /* Check for bad arguments */
  if ( ( (old_type == MPI_DATATYPE_NULL) && (errno = MPI_ERR_TYPE) )  ||
	   ( (old_type == (*newtype))        && (errno = MPI_ERR_TYPE) )  ||
	   ( (count   <= 0)                  && (errno = MPI_ERR_COUNT) ) ||
	   ( (blocklen <= 0)                 && (errno = MPI_ERR_ARG) )   )
	return MPIR_ERROR( MPI_COMM_NULL, errno,
					  "Error in MPI_TYPE_HVECTOR" );
	
  /* Handle the case where blocklen & stride make a contiguous type */
  if ( ((blocklen * old_type->extent) == stride) ||
	   (count                         == 1) )
	return MPI_Type_contiguous ( count * blocklen, old_type, newtype );

  /* Create and fill in the datatype */
  dteptr = (*newtype) = (MPI_Datatype) SBalloc( MPIR_dtes );
  dteptr->dte_type    = MPIR_HVECTOR;
  dteptr->committed   = MPIR_FALSE;
  dteptr->basic       = MPIR_FALSE;
  dteptr->is_contig   = MPIR_FALSE;
  dteptr->ref_count   = 1;
  dteptr->align       = old_type->align;
  dteptr->elements    = count * blocklen * old_type->elements;
  dteptr->stride      = stride;
  dteptr->blocklen    = blocklen;
  dteptr->old_type    = (MPI_Datatype)MPIR_Type_dup (old_type);
  dteptr->count       = count;

  /* Set the upper/lower bounds and the extent and size */
  dteptr->lb          = old_type->lb;
  dteptr->extent      = ((count-1) * stride) + (blocklen * old_type->extent);
  dteptr->ub          = dteptr->lb + dteptr->extent;
  dteptr->size        = count * blocklen * dteptr->old_type->size;
  
  return (errno);
}
