/*
 *  $Id: type_ind.c,v 1.1 1994/04/18 15:47:07 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include "mpi.h"
#include "mpisys.h"

#ifndef MPIR_TRUE
#define MPIR_TRUE  1
#define MPIR_FALSE 0
#endif


/*@
    MPI_Type_indexed - Create an indexed datatype
@*/
int MPI_Type_indexed( count, blocklens, indices, old_type, newtype )
int           count;
int 	      blocklens[];
int 	      indices[];
MPI_Datatype  old_type;
MPI_Datatype *newtype;
{
  MPIR_DATATYPE  *dteptr;
  MPI_Aint ub, high, low;
  int      i, errno = MPI_SUCCESS;

  /* Check for bad arguments */
  if ( ( (old_type == MPI_DATATYPE_NULL) && (errno = MPI_ERR_TYPE) )  ||
	   ( (old_type == (*newtype))        && (errno = MPI_ERR_TYPE) )  ||
	   ( (count    <= 0)                 && (errno = MPI_ERR_COUNT) ) )
	return MPIR_ERROR( MPI_COMM_NULL, errno,
					  "Error in MPI_TYPE_INDEXED" );
	
  /* Create and fill in the datatype */
  dteptr = (*newtype) = (MPI_Datatype) SBalloc( MPIR_dtes );
  dteptr->dte_type    = MPIR_INDEXED;
  dteptr->committed   = MPIR_FALSE;
  dteptr->basic       = MPIR_FALSE;
  dteptr->is_contig   = MPIR_FALSE;
  dteptr->ref_count   = 1;
  dteptr->align       = old_type->align;
  dteptr->old_type    = (MPI_Datatype)MPIR_Type_dup (old_type);
  dteptr->count       = count;
  dteptr->elements    = 0;

  /* Create indices and blocklens arrays and fill them */
  dteptr->indices     = ( int * ) MALLOC( count * sizeof( int ) );
  dteptr->blocklens   = ( int * ) MALLOC( count * sizeof( int ) );
  low                 = indices[0];
  high                = (indices[0]   * old_type->extent) + 
	                    (blocklens[0] * old_type->extent);
  for (i = 0; i < count; i++)  {
	dteptr->indices[i]    = indices[i] * old_type->extent;
	dteptr->blocklens[i]  = blocklens[i];
	ub = (dteptr->indices[i]   * old_type->extent) +
	     (dteptr->blocklens[i] * old_type->extent) ;
	if ( high < ub )
	  high = ub;
	low  = (low  > dteptr->indices[i]) ? dteptr->indices[i] : low;
	dteptr->elements     += blocklens[i];
  }

  /* Set the upper/lower bounds and the extent and size */
  dteptr->lb          = low * old_type->extent;
  dteptr->ub          = high;
  dteptr->extent      = dteptr->ub - dteptr->lb;
  dteptr->size        = dteptr->elements * old_type->size;
  dteptr->elements   *= old_type->elements;
  
  return (errno);
}
