/*
 *  $Id: waitany.c,v 1.3 1994/04/13 13:22:57 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */


#ifndef lint
static char vcid[] = "$Id: waitany.c,v 1.3 1994/04/13 13:22:57 gropp Exp $";
#endif /* lint */
#include "mpi.h"
#include "mpisys.h"

/*@
    MPI_Waitany - Wait for any specified send or receive to complete

Input Parameters:
. count - list length (integer) 
. array_of_requests - array of requests (array of handles) 

Output Parameters:
. index - index of handle for operation that completed (integer).  In the
range 0 to count-1.
. status - status object (Status) 
@*/
int MPI_Waitany(count, array_of_requests, index, status )
int         count;
MPI_Request array_of_requests[];
int         *index;
MPI_Status  *status;
{
    int i;
    
    while (1) {
	for (i = 0; i < count; i++)
	    {
	    /* Skip over null handle.  We need this for handles generated
	       when MPI_PROC_NULL is the source or destination of an 
	       operation */
	    if (!array_of_requests[i]) continue;
	    if (array_of_requests[i]->chandle.completed) {
		*index = i;
		if ( array_of_requests[i]->type == MPIR_RECV ) {
		    MPID_complete_recv( &array_of_requests[i]->rhandle );
		    
		    status->MPI_SOURCE = array_of_requests[i]->rhandle.source;
		    status->MPI_TAG    = array_of_requests[i]->rhandle.tag;
		    status->count  = array_of_requests[i]->rhandle.actcount;  
		    status->count_in_bytes =
			             array_of_requests[i]->rhandle.totallen;
		    }
		else
		    MPID_complete_send( &array_of_requests[i]->shandle, 
				        status );

	        /* Spec requires, but... */
		MPI_Request_free( &array_of_requests[i] ); 
		array_of_requests[i]    = NULL;
		return MPI_SUCCESS;
		}
	    }
	MPID_check_device( 1 );
	}
}

