/*
 *  $Id: map_cart.c,v 1.5 1994/03/29 22:25:49 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include <mpi.h>

/*@

MPI_Cart_map - Map process to cart topology information 

Input Parameters:
. comm - input communicator (handle) 
. ndims - number of dimensions of cartesian structure (integer) 
. dims - integer array of size  ndims specifying the number of processes in each coordinate direction 
. periods - logical array of size  ndims specifying the periodicity specification in each coordinate direction 
Output Parameter:
. newrank - reordered rank of the calling process; MPI_UNDEFINED if calling process does not belong to grid (integer) 

@*/
int MPI_Cart_map ( comm_old, ndims, dims, periods, newrank )
MPI_Comm comm_old;
int      ndims;
int     *dims;
int     *periods;
int     *newrank;
{
  int i;
  int nranks = 1;
  int rank;
  int errno = MPI_SUCCESS;

  /* Check for valid arguments */
  if (((comm_old == (MPI_Comm)0) && (errno = MPI_ERR_COMM_NULL)) ||
      ((ndims < 1)               && (errno = MPI_ERR_DIMS))      ||
      ((newrank == (int *)0)     && (errno = MPI_ERR_NULL))      ||
      ((dims == (int *)0)        && (errno = MPI_ERR_NULL))       )
    return MPIR_ERROR( comm_old, errno, "Error in MPI_CART_MAP" );
  
  /* Determine number of processes needed for topology */
  for ( i=0; i<ndims; i++ )
    nranks *= dims[i];

  /* Am I in this range? */
  MPI_Comm_rank ( comm_old, &rank );
  if ( rank < nranks )
    (*newrank) = rank;
  else
    (*newrank) = MPI_UNDEFINED;

  return (errno);
}
