/*
 *  $Id: recv.c,v 1.12 1994/04/18 21:33:57 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */


#ifndef lint
static char vcid[] = "$Id: recv.c,v 1.12 1994/04/18 21:33:57 gropp Exp $";
#endif /* lint */

#include "mpi.h"

/*@
    MPI_Recv - Basic receive

Output Parameter:
. buf - initial address of receive buffer (choice) 
. status - status object (Status) 

Input Parameters:
. count - number of elements in receive buffer (integer) 
. datatype - datatype of each receive buffer element (handle) 
. source - rank of source (integer) 
. tag - message tag (integer) 
. comm - communicator (handle) 

@*/
int MPI_Recv( buf, count, datatype, source, tag, comm, status )
void             *buf;
int              count, source, tag;
MPI_Datatype     datatype;
MPI_Comm         comm;
MPI_Status       *status;
{
    MPI_Request request;
    MPIR_RHANDLE rhandle;
    int         err, is_available;

    /* 
       Because this is a very common routine, we show how it can be
       optimized to be run "inline"; In addition, this lets us exploit
       features in the ADI to simplify the execution of blocking receive
       calls.
     */
    if (source != MPI_PROC_NULL)
    {
        request = (MPIR_HANDLE *)&rhandle;
	request->type	    = MPIR_RECV;
	if (source == MPI_ANY_SOURCE)
	    rhandle.source  = source;
	else
	    rhandle.source  = comm->group->lrank_to_grank[source];
	rhandle.tag	    = tag;
	rhandle.contextid   = comm->recv_context;
	rhandle.datatype    = datatype;
	rhandle.bufadd	    = buf;
	rhandle.count	    = count;
	rhandle.completed   = NO;
	MPID_alloc_recv_handle(&(rhandle.dev_rhandle));
	if (rhandle.datatype->is_contig) {
	    rhandle.dev_rhandle.start		= buf;
            rhandle.dev_rhandle.bytes_as_contig	= count * datatype->extent;
	    rhandle.bufpos                      = 0;
	    }
#ifdef MPID_RETURN_PACKED
	else {
	    if (err = 
	MPIR_SetupUnPackMessage( buf, count, datatype, source, request )) {
		MPIR_ERROR( comm, err, "Could not pack message in MPI_RECV" );
		}
	    }
#else
	else 
	    rhandle.dev_rhandle.start = 0;
#endif
#ifdef MPID_blocking_recv
	/* If ADI supports blocking receive directly, use that */
	MPID_blocking_recv( &rhandle );
#else
        /* ADI will handle queueing of MPIR recv handle in posted-recv queue */
	MPID_set_recv_is_nonblocking( &(rhandle.dev_rhandle), 1 );
	MPID_post_recv( &rhandle, &(is_available) );
	while ( request->rhandle.completed == NO ) {
	    MPID_check_device( MPID_BLOCKING );
	    }
	MPID_complete_recv( &rhandle );
#endif
	status->MPI_SOURCE     = rhandle.source;
	status->MPI_TAG        = rhandle.tag;
	status->count	       = rhandle.actcount;  
	status->count_in_bytes = rhandle.totallen;
#ifdef MPID_RETURN_PACKED
	if (rhandle.bufpos) 
	    err = MPIR_UnPackMessage( buf, count, datatype, source, request );
#endif
    }
    return MPI_SUCCESS;
}

